import pytest
from typing import Union, List, cast

import pyteal as pt

avm4Options = pt.CompileOptions(version=4)
avm5Options = pt.CompileOptions(version=5)
avm7Options = pt.CompileOptions(version=7)

curve_options_map = {
    pt.EcdsaCurve.Secp256k1: avm5Options,
    pt.EcdsaCurve.Secp256r1: avm7Options,
}


@pytest.mark.parametrize("curve", [pt.EcdsaCurve.Secp256k1, pt.EcdsaCurve.Secp256r1])
def test_ecdsa_decompress(curve: pt.EcdsaCurve):
    compressed_pubkey = pt.Bytes("XY")
    pubkey = pt.EcdsaDecompress(curve, compressed_pubkey)
    assert pubkey.type_of() == pt.TealType.none

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(compressed_pubkey, pt.Op.byte, '"XY"'),
            pt.TealOp(pubkey, pt.Op.ecdsa_pk_decompress, curve.arg_name),
            pt.TealOp(
                pubkey.output_slots[1].store(), pt.Op.store, pubkey.output_slots[1]
            ),
            pt.TealOp(
                pubkey.output_slots[0].store(), pt.Op.store, pubkey.output_slots[0]
            ),
        ]
    )

    actual, _ = pubkey.__teal__(curve_options_map[curve])
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected

    # compile without errors this is necessary so assembly is also tested
    pt.compileTeal(
        pt.Seq(pubkey, pt.Approve()), pt.Mode.Application, version=curve.min_version
    )

    with pytest.raises(pt.TealInputError):
        pt.compileTeal(
            pt.Seq(pubkey, pt.Approve()),
            pt.Mode.Application,
            version=curve.min_version - 1,
        )


@pytest.mark.parametrize("curve", [pt.EcdsaCurve.Secp256k1, pt.EcdsaCurve.Secp256r1])
def test_ecdsa_recover(curve: pt.EcdsaCurve):
    if curve != pt.EcdsaCurve.Secp256k1:
        with pytest.raises(pt.TealInputError):
            pt.EcdsaRecover(
                curve, pt.Bytes("data"), pt.Int(1), pt.Bytes("sigA"), pt.Bytes("sigB")
            )
    else:
        args = [pt.Bytes("data"), pt.Int(1), pt.Bytes("sigA"), pt.Bytes("sigB")]
        pubkey = pt.EcdsaRecover(curve, args[0], args[1], args[2], args[3])
        assert pubkey.type_of() == pt.TealType.none

        expected = pt.TealSimpleBlock(
            [
                pt.TealOp(args[0], pt.Op.byte, '"data"'),
                pt.TealOp(args[1], pt.Op.int, 1),
                pt.TealOp(args[2], pt.Op.byte, '"sigA"'),
                pt.TealOp(args[3], pt.Op.byte, '"sigB"'),
                pt.TealOp(pubkey, pt.Op.ecdsa_pk_recover, curve.arg_name),
                pt.TealOp(
                    pubkey.output_slots[1].store(), pt.Op.store, pubkey.output_slots[1]
                ),
                pt.TealOp(
                    pubkey.output_slots[0].store(), pt.Op.store, pubkey.output_slots[0]
                ),
            ]
        )

        actual, _ = pubkey.__teal__(curve_options_map[curve])
        actual.addIncoming()
        actual = pt.TealBlock.NormalizeBlocks(actual)

        with pt.TealComponent.Context.ignoreExprEquality():
            assert actual == expected

        # compile without errors this is necessary so assembly is also tested
        pt.compileTeal(
            pt.Seq(pubkey, pt.Approve()), pt.Mode.Application, version=curve.min_version
        )

        with pytest.raises(pt.TealInputError):
            pt.compileTeal(
                pt.Seq(pubkey, pt.Approve()),
                pt.Mode.Application,
                version=curve.min_version - 1,
            )


@pytest.mark.parametrize("curve", [pt.EcdsaCurve.Secp256k1, pt.EcdsaCurve.Secp256r1])
def test_ecdsa_verify_basic(curve: pt.EcdsaCurve):
    args = [pt.Bytes("data"), pt.Bytes("sigA"), pt.Bytes("sigB")]
    pubkey = (pt.Bytes("X"), pt.Bytes("Y"))
    expr = pt.EcdsaVerify(curve, args[0], args[1], args[2], pubkey)
    assert expr.type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.byte, '"data"'),
            pt.TealOp(args[1], pt.Op.byte, '"sigA"'),
            pt.TealOp(args[2], pt.Op.byte, '"sigB"'),
            pt.TealOp(pubkey[0], pt.Op.byte, '"X"'),
            pt.TealOp(pubkey[1], pt.Op.byte, '"Y"'),
            pt.TealOp(expr, pt.Op.ecdsa_verify, curve.arg_name),
        ]
    )

    actual, _ = expr.__teal__(curve_options_map[curve])
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected

    # compile without errors this is necessary so assembly is also tested
    pt.compileTeal(
        pt.Seq(pt.Pop(expr), pt.Approve()),
        pt.Mode.Application,
        version=curve.min_version,
    )

    with pytest.raises(pt.TealInputError):
        pt.compileTeal(
            pt.Seq(pt.Pop(expr), pt.Approve()),
            pt.Mode.Application,
            version=curve.min_version - 1,
        )


@pytest.mark.parametrize("curve", [pt.EcdsaCurve.Secp256k1, pt.EcdsaCurve.Secp256r1])
def test_ecdsa_verify_compressed_pk(curve: pt.EcdsaCurve):
    args = [pt.Bytes("data"), pt.Bytes("sigA"), pt.Bytes("sigB")]
    compressed_pubkey = pt.Bytes("XY")
    pubkey = pt.EcdsaDecompress(curve, compressed_pubkey)
    expr = pt.EcdsaVerify(curve, args[0], args[1], args[2], pubkey)
    assert expr.type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(compressed_pubkey, pt.Op.byte, '"XY"'),
            pt.TealOp(pubkey, pt.Op.ecdsa_pk_decompress, curve.arg_name),
            pt.TealOp(
                pubkey.output_slots[1].store(), pt.Op.store, pubkey.output_slots[1]
            ),
            pt.TealOp(
                pubkey.output_slots[0].store(), pt.Op.store, pubkey.output_slots[0]
            ),
            pt.TealOp(args[0], pt.Op.byte, '"data"'),
            pt.TealOp(args[1], pt.Op.byte, '"sigA"'),
            pt.TealOp(args[2], pt.Op.byte, '"sigB"'),
            pt.TealOp(
                pubkey.output_slots[0].load(), pt.Op.load, pubkey.output_slots[0]
            ),
            pt.TealOp(
                pubkey.output_slots[1].load(), pt.Op.load, pubkey.output_slots[1]
            ),
            pt.TealOp(expr, pt.Op.ecdsa_verify, curve.arg_name),
        ]
    )

    actual, _ = expr.__teal__(curve_options_map[curve])
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected

    # compile without errors this is necessary so assembly is also tested
    pt.compileTeal(
        pt.Seq(pt.Pop(expr), pt.Approve()),
        pt.Mode.Application,
        version=curve.min_version,
    )

    with pytest.raises(pt.TealInputError):
        pt.compileTeal(
            pt.Seq(pt.Pop(expr), pt.Approve()),
            pt.Mode.Application,
            version=curve.min_version - 1,
        )


def test_ecdsa_verify_recovered_pk():
    curve = pt.EcdsaCurve.Secp256k1
    args = [pt.Bytes("data"), pt.Int(1), pt.Bytes("sigA"), pt.Bytes("sigB")]
    pubkey = pt.EcdsaRecover(curve, args[0], args[1], args[2], args[3])
    expr = pt.EcdsaVerify(curve, args[0], args[2], args[3], pubkey)
    assert expr.type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.byte, '"data"'),
            pt.TealOp(args[1], pt.Op.int, 1),
            pt.TealOp(args[2], pt.Op.byte, '"sigA"'),
            pt.TealOp(args[3], pt.Op.byte, '"sigB"'),
            pt.TealOp(pubkey, pt.Op.ecdsa_pk_recover, curve.arg_name),
            pt.TealOp(
                pubkey.output_slots[1].store(), pt.Op.store, pubkey.output_slots[1]
            ),
            pt.TealOp(
                pubkey.output_slots[0].store(), pt.Op.store, pubkey.output_slots[0]
            ),
            pt.TealOp(args[0], pt.Op.byte, '"data"'),
            pt.TealOp(args[1], pt.Op.byte, '"sigA"'),
            pt.TealOp(args[2], pt.Op.byte, '"sigB"'),
            pt.TealOp(
                pubkey.output_slots[0].load(), pt.Op.load, pubkey.output_slots[0]
            ),
            pt.TealOp(
                pubkey.output_slots[1].load(), pt.Op.load, pubkey.output_slots[1]
            ),
            pt.TealOp(expr, pt.Op.ecdsa_verify, curve.arg_name),
        ]
    )

    actual, _ = expr.__teal__(curve_options_map[curve])
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected

    # compile without errors this is necessary so assembly is also tested
    pt.compileTeal(
        pt.Seq(pt.Pop(expr), pt.Approve()),
        pt.Mode.Application,
        version=curve.min_version,
    )

    with pytest.raises(pt.TealInputError):
        pt.compileTeal(
            pt.Seq(pt.Pop(expr), pt.Approve()),
            pt.Mode.Application,
            version=curve.min_version - 1,
        )


@pytest.mark.parametrize("curve", [pt.EcdsaCurve.Secp256k1, pt.EcdsaCurve.Secp256r1])
def test_ecdsa_invalid(curve: pt.EcdsaCurve):
    if curve == pt.EcdsaCurve.Secp256k1:
        with pytest.raises(pt.TealTypeError):
            args: List[Union[pt.Bytes, pt.Int]] = [
                pt.Bytes("data"),
                pt.Bytes("1"),
                pt.Bytes("sigA"),
                pt.Bytes("sigB"),
            ]
            pt.EcdsaRecover(curve, args[0], args[1], args[2], args[3])

    with pytest.raises(pt.TealTypeError):
        pt.EcdsaDecompress(curve, pt.Int(1))

    with pytest.raises(pt.TealTypeError):
        args = [pt.Bytes("data"), pt.Bytes("sigA"), pt.Bytes("sigB")]
        pubkey: Union[tuple[pt.Bytes, Union[pt.Int, pt.Bytes]], pt.MultiValue] = (
            pt.Bytes("X"),
            pt.Int(1),
        )
        pt.EcdsaVerify(curve, args[0], args[1], args[2], pubkey)

    with pytest.raises(pt.TealTypeError):
        args = [pt.Bytes("data"), pt.Int(1), pt.Bytes("sigB")]
        pubkey = (pt.Bytes("X"), pt.Bytes("Y"))
        pt.EcdsaVerify(curve, args[0], args[1], args[2], pubkey)

    with pytest.raises(pt.TealTypeError):
        args = [pt.Bytes("data"), pt.Bytes("sigA"), pt.Bytes("sigB")]
        compressed_pk = pt.Bytes("XY")
        pubkey = pt.MultiValue(
            pt.Op.ecdsa_pk_decompress,
            [pt.TealType.uint64, pt.TealType.bytes],
            immediate_args=[curve.__str__()],
            args=[compressed_pk],
        )
        pt.EcdsaVerify(curve, args[0], args[1], args[2], pubkey)

    with pytest.raises(pt.TealInputError):
        args = [pt.Bytes("data"), pt.Bytes("sigA"), pt.Bytes("sigB")]
        pubkey = (pt.Bytes("X"), pt.Bytes("Y"))
        expr = pt.EcdsaVerify(curve, args[0], args[1], args[2], pubkey)

        expr.__teal__(avm4Options)

    with pytest.raises(pt.TealTypeError):
        args = [pt.Bytes("data"), pt.Bytes("sigA"), pt.Bytes("sigB")]
        pubkey = (pt.Bytes("X"), pt.Bytes("Y"))
        expr = pt.EcdsaVerify(cast(pt.EcdsaCurve, 5), args[0], args[1], args[2], pubkey)



import pytest

import pyteal as pt

options = pt.CompileOptions(version=4)


def test_main_return():
    arg = pt.Int(1)
    expr = pt.Return(arg)
    assert expr.type_of() == pt.TealType.none
    assert expr.has_return()

    expected = pt.TealSimpleBlock(
        [pt.TealOp(arg, pt.Op.int, 1), pt.TealOp(expr, pt.Op.return_)]
    )

    actual, _ = expr.__teal__(options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected


def test_main_return_invalid():
    with pytest.raises(pt.TealCompileError):
        pt.Return(pt.Txn.receiver()).__teal__(options)

    with pytest.raises(pt.TealCompileError):
        pt.Return().__teal__(options)


def test_subroutine_return_value():
    cases = (
        (pt.TealType.uint64, pt.Int(1), pt.Op.int, 1),
        (pt.TealType.bytes, pt.Bytes("value"), pt.Op.byte, '"value"'),
        (pt.TealType.anytype, pt.Int(1), pt.Op.int, 1),
        (pt.TealType.anytype, pt.Bytes("value"), pt.Op.byte, '"value"'),
    )

    for (tealType, value, op, opValue) in cases:
        expr = pt.Return(value)

        def mySubroutine():
            return expr

        subroutine = pt.SubroutineDefinition(mySubroutine, tealType)

        assert expr.type_of() == pt.TealType.none
        assert expr.has_return()

        expected = pt.TealSimpleBlock(
            [pt.TealOp(value, op, opValue), pt.TealOp(expr, pt.Op.retsub)]
        )

        options.setSubroutine(subroutine)
        actual, _ = expr.__teal__(options)
        options.setSubroutine(None)

        actual.addIncoming()
        actual = pt.TealBlock.NormalizeBlocks(actual)

        assert actual == expected


def test_subroutine_return_value_invalid():
    cases = (
        (pt.TealType.bytes, pt.Int(1)),
        (pt.TealType.uint64, pt.Bytes("value")),
    )

    for (tealType, value) in cases:
        expr = pt.Return(value)

        def mySubroutine():
            return expr

        subroutine = pt.SubroutineDefinition(mySubroutine, tealType)

        options.setSubroutine(subroutine)
        with pytest.raises(pt.TealCompileError):
            expr.__teal__(options)
        options.setSubroutine(None)


def test_subroutine_return_none():
    expr = pt.Return()

    def mySubroutine():
        return expr

    subroutine = pt.SubroutineDefinition(mySubroutine, pt.TealType.none)

    assert expr.type_of() == pt.TealType.none
    assert expr.has_return()

    expected = pt.TealSimpleBlock([pt.TealOp(expr, pt.Op.retsub)])

    options.setSubroutine(subroutine)
    actual, _ = expr.__teal__(options)
    options.setSubroutine(None)

    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected


def test_subroutine_return_none_invalid():
    for value in (pt.Int(1), pt.Bytes("value")):
        expr = pt.Return(value)

        def mySubroutine():
            return expr

        subroutine = pt.SubroutineDefinition(mySubroutine, pt.TealType.none)

        options.setSubroutine(subroutine)
        with pytest.raises(pt.TealCompileError):
            expr.__teal__(options)
        options.setSubroutine(None)



import pytest

import pyteal as pt
from pyteal import abi
from pyteal.ast.abi.util import substring_for_decoding
from pyteal.ast.abi.type_test import ContainerType

options = pt.CompileOptions(version=5)


def test_StringTypeSpec_str():
    assert str(abi.StringTypeSpec()) == "string"


def test_StringTypeSpec_is_dynamic():
    assert (abi.StringTypeSpec()).is_dynamic()


def test_StringTypeSpec_new_instance():
    assert isinstance(abi.StringTypeSpec().new_instance(), abi.String)


def test_StringTypeSpec_eq():
    assert abi.StringTypeSpec() == abi.StringTypeSpec()

    for otherType in (
        abi.ByteTypeSpec(),
        abi.StaticArrayTypeSpec(abi.ByteTypeSpec(), 1),
        abi.DynamicArrayTypeSpec(abi.Uint8TypeSpec()),
        abi.DynamicArrayTypeSpec(abi.ByteTypeSpec()),
    ):
        assert abi.StringTypeSpec() != otherType


def test_String_encode():
    value = abi.String()
    expr = value.encode()
    assert expr.type_of() == pt.TealType.bytes
    assert expr.has_return() is False

    expected = pt.TealSimpleBlock(
        [pt.TealOp(expr, pt.Op.load, value.stored_value.slot)]
    )
    actual, _ = expr.__teal__(options)
    assert actual == expected


def test_DynamicArray_decode():
    encoded = pt.Bytes("encoded")
    stringType = abi.StringTypeSpec()
    for start_index in (None, pt.Int(1)):
        for end_index in (None, pt.Int(2)):
            for length in (None, pt.Int(3)):
                value = stringType.new_instance()

                if end_index is not None and length is not None:
                    with pytest.raises(pt.TealInputError):
                        value.decode(
                            encoded,
                            start_index=start_index,
                            end_index=end_index,
                            length=length,
                        )
                    continue

                expr = value.decode(
                    encoded, start_index=start_index, end_index=end_index, length=length
                )
                assert expr.type_of() == pt.TealType.none
                assert expr.has_return() is False

                expectedExpr = value.stored_value.store(
                    substring_for_decoding(
                        encoded,
                        start_index=start_index,
                        end_index=end_index,
                        length=length,
                    )
                )
                expected, _ = expectedExpr.__teal__(options)
                expected.addIncoming()
                expected = pt.TealBlock.NormalizeBlocks(expected)

                actual, _ = expr.__teal__(options)
                actual.addIncoming()
                actual = pt.TealBlock.NormalizeBlocks(actual)

                with pt.TealComponent.Context.ignoreExprEquality():
                    assert actual == expected


def test_String_get():
    value = abi.String()
    expr = value.get()
    assert expr.type_of() == pt.TealType.bytes
    assert expr.has_return() is False

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(expr, pt.Op.load, value.stored_value.slot),
            pt.TealOp(None, pt.Op.extract, 2, 0),
        ]
    )
    actual, _ = expr.__teal__(options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


STATIC_SET_TESTCASES: list[tuple[str | bytes | bytearray, bytes]] = [
    ("stringy", b"\x00\x07stringy"),
    ("😀", b"\x00\x04\xf0\x9f\x98\x80"),
    ("0xDEADBEEF", b"\x00\x0a0xDEADBEEF"),
    (bytes(32), b"\x00\x20" + bytes(32)),
    (b"alphabet_soup", b"\x00\x0dalphabet_soup"),
    (bytearray(b"another one"), b"\x00\x0banother one"),
]


@pytest.mark.parametrize("value_to_set, value_encoded", STATIC_SET_TESTCASES)
def test_String_set_static(value_to_set, value_encoded):
    value = abi.String()
    expr = value.set(value_to_set)
    assert expr.type_of() == pt.TealType.none
    assert not expr.has_return()

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(None, pt.Op.byte, "0x" + value_encoded.hex()),
            pt.TealOp(None, pt.Op.store, value.stored_value.slot),
        ]
    )

    actual, _ = expr.__teal__(options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def test_String_set_static_invalid():
    with pytest.raises(pt.TealInputError):
        abi.String().set(42)


def test_String_set_expr():
    for value_to_set in (pt.Bytes("hi"), pt.Bytes("base16", "0xdeadbeef")):
        value = abi.String()
        expr = value.set(value_to_set)
        assert expr.type_of() == pt.TealType.none
        assert not expr.has_return()

        value_start, value_end = value_to_set.__teal__(options)
        expected_body = pt.TealSimpleBlock(
            [
                pt.TealOp(None, pt.Op.store, value.stored_value.slot),
                pt.TealOp(None, pt.Op.load, value.stored_value.slot),
                pt.TealOp(None, pt.Op.len),
                pt.TealOp(None, pt.Op.itob),
                pt.TealOp(None, pt.Op.extract, 6, 0),
                pt.TealOp(None, pt.Op.load, value.stored_value.slot),
                pt.TealOp(None, pt.Op.concat),
                pt.TealOp(None, pt.Op.store, value.stored_value.slot),
            ]
        )
        value_end.setNextBlock(expected_body)
        expected = value_start
        expected.addIncoming()
        expected = pt.TealBlock.NormalizeBlocks(expected)

        actual, _ = expr.__teal__(options)
        actual.addIncoming()
        actual = pt.TealBlock.NormalizeBlocks(actual)

        with pt.TealComponent.Context.ignoreExprEquality():
            assert actual == expected


def test_String_set_copy():
    value = abi.String()
    other = abi.String()
    expr = value.set(other)
    assert expr.type_of() == pt.TealType.none
    assert not expr.has_return()

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(None, pt.Op.load, other.stored_value.slot),
            pt.TealOp(None, pt.Op.store, value.stored_value.slot),
        ]
    )

    actual, _ = expr.__teal__(options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected

    with pytest.raises(pt.TealInputError):
        value.set(abi.Address())


def test_String_set_computed():
    bv = pt.Bytes("base16", "0x0004DEADBEEF")
    computed_value = ContainerType(abi.StringTypeSpec(), bv)

    value = abi.String()
    expr = value.set(computed_value)
    assert expr.type_of() == pt.TealType.none
    assert not expr.has_return()

    _, byte_ops = bv.__teal__(options)
    expected = pt.TealSimpleBlock(
        [
            byte_ops.ops[0],
            pt.TealOp(None, pt.Op.store, value.stored_value.slot),
        ]
    )

    actual, _ = expr.__teal__(options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected

    with pytest.raises(pt.TealInputError):
        value.set(ContainerType(abi.ByteTypeSpec(), pt.Int(0x01)))



from pathlib import Path
import pytest
import json

import pyteal as pt


def test_abi_algobank():
    from examples.application.abi.algobank import (
        approval_program,
        clear_state_program,
        contract,
    )

    target_dir = Path.cwd() / "examples" / "application" / "abi"

    with open(
        target_dir / "algobank_approval.teal", "r"
    ) as expected_approval_program_file:
        expected_approval_program = "".join(
            expected_approval_program_file.readlines()
        ).strip()
        assert approval_program == expected_approval_program

    with open(
        target_dir / "algobank_clear_state.teal", "r"
    ) as expected_clear_state_program_file:
        expected_clear_state_program = "".join(
            expected_clear_state_program_file.readlines()
        ).strip()
        assert clear_state_program == expected_clear_state_program

    with open(target_dir / "algobank.json", "r") as expected_contract_file:
        expected_contract = json.load(expected_contract_file)
        assert contract.dictify() == expected_contract


def test_basic_bank():
    from examples.signature.basic import bank_for_account

    program = bank_for_account(
        "ZZAF5ARA4MEC5PVDOP64JM5O5MQST63Q2KOY2FLYFLXXD3PFSNJJBYAFZM"
    )

    target_path = Path.cwd() / "examples" / "signature" / "basic.teal"

    with open(target_path, "r") as target_file:
        target = "".join(target_file.readlines()).strip()
        assert pt.compileTeal(program, mode=pt.Mode.Signature, version=3) == target


def test_atomic_swap():
    from examples.signature.atomic_swap import htlc

    program = htlc()

    target_path = Path.cwd() / "examples" / "signature" / "atomic_swap.teal"

    with open(target_path, "r") as target_file:
        target = "".join(target_file.readlines()).strip()
        assert pt.compileTeal(program, mode=pt.Mode.Signature, version=2) == target


def test_periodic_payment():
    from examples.signature.periodic_payment import periodic_payment

    program = periodic_payment()

    target_path = Path.cwd() / "examples" / "signature" / "periodic_payment.teal"
    with open(target_path, "r") as target_file:
        target = "".join(target_file.readlines()).strip()
        assert pt.compileTeal(program, mode=pt.Mode.Signature, version=2) == target


def test_split():
    from examples.signature.split import split

    program = split()

    target_path = Path.cwd() / "examples" / "signature" / "split.teal"

    with open(target_path, "r") as target_file:
        target = "".join(target_file.readlines()).strip()
        assert pt.compileTeal(program, mode=pt.Mode.Signature, version=2) == target


def test_dutch_auction():
    from examples.signature.dutch_auction import dutch_auction

    program = dutch_auction()

    target_path = Path.cwd() / "examples" / "signature" / "dutch_auction.teal"

    with open(target_path, "r") as target_file:
        target = "".join(target_file.readlines()).strip()
        assert pt.compileTeal(program, mode=pt.Mode.Signature, version=2) == target


def test_recurring_swap():
    from examples.signature.recurring_swap import recurring_swap

    program = recurring_swap()

    target_path = Path.cwd() / "examples" / "signature" / "recurring_swap.teal"

    with open(target_path, "r") as target_file:
        target = "".join(target_file.readlines()).strip()
        assert pt.compileTeal(program, mode=pt.Mode.Signature, version=2) == target


def test_asset():
    from examples.application.asset import approval_program, clear_state_program

    approval = approval_program()
    clear_state = clear_state_program()

    # only checking for successful compilation for now
    pt.compileTeal(approval, mode=pt.Mode.Application, version=2)
    pt.compileTeal(clear_state, mode=pt.Mode.Application, version=2)


def test_security_token():
    from examples.application.security_token import (
        approval_program,
        clear_state_program,
    )

    approval = approval_program()
    clear_state = clear_state_program()

    # only checking for successful compilation for now
    pt.compileTeal(approval, mode=pt.Mode.Application, version=2)
    pt.compileTeal(clear_state, mode=pt.Mode.Application, version=2)


def test_vote():
    from examples.application.vote import approval_program, clear_state_program

    approval = approval_program()
    clear_state = clear_state_program()

    # only checking for successful compilation for now
    pt.compileTeal(approval, mode=pt.Mode.Application, version=2)
    pt.compileTeal(clear_state, mode=pt.Mode.Application, version=2)


def test_cond():
    cond1 = pt.Txn.fee() < pt.Int(2000)
    cond2 = pt.Txn.amount() > pt.Int(5000)
    cond3 = pt.Txn.receiver() == pt.Txn.sender()
    core = pt.Cond(
        [pt.Global.group_size() == pt.Int(2), cond1],
        [pt.Global.group_size() == pt.Int(3), cond2],
        [pt.Global.group_size() == pt.Int(4), cond3],
    )
    pt.compileTeal(core, mode=pt.Mode.Signature, version=2)


@pytest.mark.timeout(2)
def test_many_ifs():
    """
    Test with many pt.If statements to trigger potential corner cases in code generation.
    Previous versions of PyTeal took an exponential time to generate the TEAL code for this PyTEAL.
    """

    sv = pt.ScratchVar(pt.TealType.uint64)
    s = pt.Seq(
        [
            pt.If(
                pt.Int(3 * i) == pt.Int(3 * i),
                sv.store(pt.Int(3 * i + 1)),
                sv.store(pt.Int(3 * i + 2)),
            )
            for i in range(30)
        ]
        + [pt.Return(sv.load())]
    )

    pt.compileTeal(s, mode=pt.Mode.Signature, version=2)



import pytest

import pyteal as pt

options = pt.CompileOptions()


def test_scratch_slot():
    slot = pt.ScratchSlot()
    assert slot == slot
    assert slot.__hash__() == slot.__hash__()
    assert slot != pt.ScratchSlot()

    with pt.TealComponent.Context.ignoreExprEquality():
        assert (
            slot.store().__teal__(options)[0]
            == pt.ScratchStackStore(slot).__teal__(options)[0]
        )
        assert (
            slot.store(pt.Int(1)).__teal__(options)[0]
            == pt.ScratchStore(slot, pt.Int(1)).__teal__(options)[0]
        )

        assert slot.load().type_of() == pt.TealType.anytype
        assert slot.load(pt.TealType.uint64).type_of() == pt.TealType.uint64
        assert (
            slot.load().__teal__(options)[0]
            == pt.ScratchLoad(slot).__teal__(options)[0]
        )


def test_scratch_load_default():
    slot = pt.ScratchSlot()
    expr = pt.ScratchLoad(slot)
    assert expr.type_of() == pt.TealType.anytype

    expected = pt.TealSimpleBlock([pt.TealOp(expr, pt.Op.load, slot)])

    actual, _ = expr.__teal__(options)

    assert actual == expected


def test_scratch_load_index_expression():
    expr = pt.ScratchLoad(slot=None, index_expression=pt.Int(1337))
    assert expr.type_of() == pt.TealType.anytype

    expected = pt.TealSimpleBlock([pt.TealOp(pt.Int(1337), pt.Op.int, 1337)])
    expected.setNextBlock(pt.TealSimpleBlock([pt.TealOp(None, pt.Op.loads)]))

    actual, _ = expr.__teal__(options)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def test_scratch_load_type():
    for type in (pt.TealType.uint64, pt.TealType.bytes, pt.TealType.anytype):
        slot = pt.ScratchSlot()
        expr = pt.ScratchLoad(slot, type)
        assert expr.type_of() == type

        expected = pt.TealSimpleBlock([pt.TealOp(expr, pt.Op.load, slot)])

        actual, _ = expr.__teal__(options)

        assert actual == expected


def test_scratch_store():
    for value in (
        pt.Int(1),
        pt.Bytes("test"),
        pt.App.globalGet(pt.Bytes("key")),
        pt.If(pt.Int(1), pt.Int(2), pt.Int(3)),
    ):
        slot = pt.ScratchSlot()
        expr = pt.ScratchStore(slot, value)
        assert expr.type_of() == pt.TealType.none

        expected, valueEnd = value.__teal__(options)
        storeBlock = pt.TealSimpleBlock([pt.TealOp(expr, pt.Op.store, slot)])
        valueEnd.setNextBlock(storeBlock)

        actual, _ = expr.__teal__(options)

        assert actual == expected


def test_scratch_store_index_expression():
    for value in (
        pt.Int(1),
        pt.Bytes("test"),
        pt.App.globalGet(pt.Bytes("key")),
        pt.If(pt.Int(1), pt.Int(2), pt.Int(3)),
    ):
        expr = pt.ScratchStore(slot=None, value=value, index_expression=pt.Int(1337))
        assert expr.type_of() == pt.TealType.none

        expected = pt.TealSimpleBlock([pt.TealOp(None, pt.Op.int, 1337)])
        valueStart, valueEnd = value.__teal__(options)
        expected.setNextBlock(valueStart)

        storeBlock = pt.TealSimpleBlock([pt.TealOp(expr, pt.Op.stores)])
        valueEnd.setNextBlock(storeBlock)

        actual, _ = expr.__teal__(options)

        with pt.TealComponent.Context.ignoreExprEquality():
            assert actual == expected


def test_scratch_stack_store():
    slot = pt.ScratchSlot()
    expr = pt.ScratchStackStore(slot)
    assert expr.type_of() == pt.TealType.none

    expected = pt.TealSimpleBlock([pt.TealOp(expr, pt.Op.store, slot)])

    actual, _ = expr.__teal__(options)

    assert actual == expected


def test_scratch_assign_id():
    slot = pt.ScratchSlot(255)
    expr = pt.ScratchStackStore(slot)
    assert expr.type_of() == pt.TealType.none

    expected = pt.TealSimpleBlock([pt.TealOp(expr, pt.Op.store, slot)])

    actual, _ = expr.__teal__(options)

    assert actual == expected


def test_scratch_assign_id_invalid():
    with pytest.raises(pt.TealInputError):
        pt.ScratchSlot(-1)

    with pytest.raises(pt.TealInputError):
        pt.ScratchSlot(pt.NUM_SLOTS)


def test_scratch_index():
    slot = pt.ScratchSlot()

    index = pt.ScratchIndex(slot)
    assert index.slot is slot

    assert str(index) == "(ScratchIndex " + str(slot) + ")"

    assert index.type_of() == pt.TealType.uint64

    assert not index.has_return()

    expected = pt.TealSimpleBlock([pt.TealOp(index, pt.Op.int, slot)])
    actual, _ = index.__teal__(options)

    assert actual == expected



from typing import TypeVar, Generic, Callable, Final, cast
from abc import ABC, abstractmethod

from pyteal.errors import TealInputError
from pyteal.types import TealType
from pyteal.ast.expr import Expr
from pyteal.ast.scratchvar import ScratchVar
from pyteal.ast.seq import Seq


class TypeSpec(ABC):
    """TypeSpec represents a specification for an ABI type.

    Essentially this is a factory that can produce specific instances of ABI types.
    """

    @abstractmethod
    def new_instance(self) -> "BaseType":
        """Create a new instance of the specified type."""
        pass

    @abstractmethod
    def annotation_type(self) -> "type[BaseType]":
        """Get the annotation type associated with this spec"""
        pass

    @abstractmethod
    def is_dynamic(self) -> bool:
        """Check if this ABI type is dynamic.

        If a type is dynamic, the length of its encoding depends on its value. Otherwise, the type
        is considered static (not dynamic).
        """
        pass

    @abstractmethod
    def byte_length_static(self) -> int:
        """Get the byte length of this ABI type's encoding. Only valid for static types."""
        pass

    @abstractmethod
    def storage_type(self) -> TealType:
        """Get the TealType that the underlying ScratchVar should hold for this type."""
        pass

    @abstractmethod
    def __eq__(self, other: object) -> bool:
        """Check if this type is considered equal to another ABI type.

        Args:
            other: The object to compare to. If this is not a TypeSpec, this method will never
                return true.

        Returns:
            True if and only if self and other represent the same ABI type.
        """
        pass

    @abstractmethod
    def __str__(self) -> str:
        """Get the string representation of this ABI type, used for creating method signatures."""
        pass


TypeSpec.__module__ = "pyteal.abi"


class BaseType(ABC):
    """The abstract base class for all ABI type instances.

    The value of the type is contained in a unique ScratchVar that only this instance has access to.
    As a result, the value of an ABI type is mutable and can be efficiently referenced multiple
    times without needing to recompute it.
    """

    def __init__(self, spec: TypeSpec) -> None:
        """Create a new BaseType."""
        super().__init__()
        self._type_spec: Final = spec
        self.stored_value: Final = ScratchVar(spec.storage_type())

    def type_spec(self) -> TypeSpec:
        """Get the TypeSpec for this ABI type instance."""
        return self._type_spec

    @abstractmethod
    def encode(self) -> Expr:
        """Encode this ABI type to a byte string.

        Returns:
            A PyTeal expression that encodes this type to a byte string.
        """
        pass

    @abstractmethod
    def decode(
        self,
        encoded: Expr,
        *,
        start_index: Expr = None,
        end_index: Expr = None,
        length: Expr = None,
    ) -> Expr:
        """Decode a substring of the passed in encoded string and set it as this type's value.

        The arguments to this function are means to be as flexible as possible for the caller.
        Multiple types of substrings can be specified based on the arguments, as listed below:

        * Entire string: if start_index, end_index, and length are all None, the entire encoded string
          is decoded.
        * Prefix: if start_index is None and one of end_index or length is provided, a prefix of the
          encoded string is decoded. The range is 0 through end_index or length (they are equivalent).
        * Suffix: if start_index is provided and end_index and length are None, a suffix of the encoded
          string is decoded. The range is start_index through the end of the string.
        * Substring specified with end_index: if start_index and end_index are provided and length is
          None, a substring of the encoded string is decoded. The range is start_index through
          end_index.
        * Substring specified with length: if start_index and length are provided and end_index is
          None, a substring of the encoded string is decoded. The range is start_index through
          start_index+length.

        Args:
            encoded: An expression containing the bytes to decode. Must evaluate to TealType.bytes.
            start_index (optional): An expression containing the index to start decoding. Must
                evaluate to TealType.uint64. Defaults to None.
            end_index (optional): An expression containing the index to stop decoding. Must evaluate
                to TealType.uint64. Defaults to None.
            length (optional): An expression containing the length of the substring to decode. Must
                evaluate to TealType.uint64. Defaults to None.

        Returns:
            An expression that performs the necessary steps in order to decode the given string into
            a value.
        """
        pass

    def _set_with_computed_type(self, value: "ComputedValue[BaseType]") -> Expr:
        target_type_spec = value.produced_type_spec()
        if self.type_spec() != target_type_spec:
            raise TealInputError(
                f"Cannot set {self.type_spec()} with ComputedType of {target_type_spec}"
            )
        return value.store_into(self)

    def __str__(self) -> str:
        return str(self.type_spec())


BaseType.__module__ = "pyteal.abi"

T_co = TypeVar("T_co", bound=BaseType, covariant=True)


class ComputedValue(ABC, Generic[T_co]):
    """Represents an ABI Type whose value must be computed by an expression."""

    @abstractmethod
    def produced_type_spec(self) -> TypeSpec:
        """Get the ABI TypeSpec that this object produces."""
        pass

    @abstractmethod
    def store_into(self, output: T_co) -> Expr:  # type: ignore[misc]
        """Compute the value and store it into an existing ABI type instance.

        NOTE: If you call this method multiple times, the computation to determine the value will be
        repeated each time. For this reason, it is recommended to only issue a single call to either
        :code:`store_into` or :code:`use`.

        Args:
            output: The object where the computed value will be stored. This object must have the
                same type as this class's produced type.

        Returns:
            An expression which stores the computed value represented by this class into the output
            object.
        """
        pass

    def use(self, action: Callable[[T_co], Expr]) -> Expr:
        """Compute the value and pass it to a callable expression.

        NOTE: If you call this method multiple times, the computation to determine the value will be
        repeated each time. For this reason, it is recommended to only issue a single call to either
        :code:`store_into` or :code:`use`.

        Args:
            action: A callable object that will receive an instance of this class's produced type
                with the computed value. The callable object may use that value as it sees fit, but
                it must return an Expr to be included in the program's AST.

        Returns:
            An expression which contains the returned expression from invoking `action` with the
            computed value.
        """
        newInstance = cast(T_co, self.produced_type_spec().new_instance())
        return Seq(self.store_into(newInstance), action(newInstance))


ComputedValue.__module__ = "pyteal.abi"


class ReturnedValue(ComputedValue):
    def __init__(self, type_spec: TypeSpec, computation_expr: Expr):
        from pyteal.ast.subroutine import SubroutineCall

        self.type_spec = type_spec
        if not isinstance(computation_expr, SubroutineCall):
            raise TealInputError(
                f"Expecting computation_expr to be SubroutineCall but get {type(computation_expr)}"
            )
        self.computation = computation_expr

    def produced_type_spec(self) -> TypeSpec:
        return self.type_spec

    def store_into(self, output: BaseType) -> Expr:
        if output.type_spec() != self.produced_type_spec():
            raise TealInputError(
                f"expected type_spec {self.produced_type_spec()} but get {output.type_spec()}"
            )

        declaration = self.computation.subroutine.get_declaration()

        if declaration.deferred_expr is None:
            raise TealInputError(
                "ABI return subroutine must have deferred_expr to be not-None."
            )
        if declaration.deferred_expr.type_of() != output.type_spec().storage_type():
            raise TealInputError(
                f"ABI return subroutine deferred_expr is expected to be typed {output.type_spec().storage_type()}, "
                f"but has type {declaration.deferred_expr.type_of()}."
            )
        return output.stored_value.slot.store(self.computation)


ReturnedValue.__module__ = "pyteal.abi"



from typing import List, Dict, DefaultDict, Optional
from collections import defaultdict

from pyteal.ast import SubroutineDefinition
from pyteal.ir import (
    Op,
    TealOp,
    TealLabel,
    TealComponent,
    TealBlock,
    TealSimpleBlock,
    TealConditionalBlock,
    LabelReference,
)
from pyteal.errors import TealInternalError


def flattenBlocks(blocks: List[TealBlock]) -> List[TealComponent]:
    """Lowers a list of TealBlocks into a list of TealComponents.

    Args:
        blocks: The blocks to lower.
    """
    codeblocks = []
    references: DefaultDict[int, int] = defaultdict(int)

    labelRefs: Dict[int, LabelReference] = dict()

    def indexToLabel(index: int) -> LabelReference:
        if index not in labelRefs:
            labelRefs[index] = LabelReference("l{}".format(index))
        return labelRefs[index]

    def blockIndexByReference(block: TealBlock) -> int:
        for i, b in enumerate(blocks):
            if block is b:
                return i
        raise ValueError("Block not present in list: {}".format(block))

    for i, block in enumerate(blocks):
        code = list(block.ops)
        codeblocks.append(code)
        if block.isTerminal():
            continue

        if type(block) is TealSimpleBlock:
            assert block.nextBlock is not None

            nextIndex = blockIndexByReference(block.nextBlock)

            if nextIndex != i + 1:
                references[nextIndex] += 1
                code.append(TealOp(None, Op.b, indexToLabel(nextIndex)))

        elif type(block) is TealConditionalBlock:
            assert block.trueBlock is not None
            assert block.falseBlock is not None

            trueIndex = blockIndexByReference(block.trueBlock)
            falseIndex = blockIndexByReference(block.falseBlock)

            if falseIndex == i + 1:
                references[trueIndex] += 1
                code.append(TealOp(None, Op.bnz, indexToLabel(trueIndex)))
                continue

            if trueIndex == i + 1:
                references[falseIndex] += 1
                code.append(TealOp(None, Op.bz, indexToLabel(falseIndex)))
                continue

            references[trueIndex] += 1
            code.append(TealOp(None, Op.bnz, indexToLabel(trueIndex)))

            references[falseIndex] += 1
            code.append(TealOp(None, Op.b, indexToLabel(falseIndex)))
        else:
            raise TealInternalError("Unrecognized block type: {}".format(type(block)))

    teal: List[TealComponent] = []
    for i, code in enumerate(codeblocks):
        if references[i] != 0:
            teal.append(TealLabel(None, indexToLabel(i)))
        teal += code

    return teal


def flattenSubroutines(
    subroutineMapping: Dict[Optional[SubroutineDefinition], List[TealComponent]],
    subroutineToLabel: Dict[SubroutineDefinition, str],
) -> List[TealComponent]:
    """Combines each subroutine's list of TealComponents into a single list of TealComponents that
    represents the entire program.

    Args:
        subroutineMapping: A dictionary containing a list of TealComponents for every subroutine in
            a program. The key None is taken to indicate the main program routine.
        subroutineToLabel: An ordered dictionary which resolves each subroutine to a string label.

    Returns:
        A single list of TealComponents representing the entire program.
    """
    combinedOps: List[TealComponent] = []

    # By default all branch labels in each subroutine will start from "l0". To
    # make each subroutine have unique labels, we prefix "main_" to the ones
    # from the main routine, and "subX_" (the subroutine label) to the
    # ones from each subroutine

    mainRoutine = subroutineMapping[None]
    for stmt in mainRoutine:
        if isinstance(stmt, TealLabel):
            stmt.getLabelRef().addPrefix("main_")
    combinedOps += mainRoutine

    for subroutine, label in subroutineToLabel.items():
        comment = subroutine.name()
        labelPrefix = label + "_"

        subroutineOps = subroutineMapping[subroutine]
        for stmt in subroutineOps:
            if isinstance(stmt, TealLabel):
                stmt.getLabelRef().addPrefix(labelPrefix)

        combinedOps.append(TealLabel(None, LabelReference(label), comment))
        combinedOps += subroutineOps

    return combinedOps



import pytest

import pyteal as pt

options = pt.CompileOptions()


def test_continue_fail():
    with pytest.raises(pt.TealCompileError):
        pt.Continue().__teal__(options)

    with pytest.raises(pt.TealCompileError):
        pt.If(pt.Int(1), pt.Continue()).__teal__(options)

    with pytest.raises(pt.TealCompileError):
        pt.Seq([pt.Continue()]).__teal__(options)

    with pytest.raises(TypeError):
        pt.Continue(pt.Int(1))


def test_continue():

    expr = pt.Continue()

    assert expr.type_of() == pt.TealType.none
    assert not expr.has_return()

    expected = pt.TealSimpleBlock([])

    options.enterLoop()
    actual, _ = expr.__teal__(options)
    breakBlocks, continueBlocks = options.exitLoop()

    assert actual == expected
    assert breakBlocks == []
    assert continueBlocks == [actual]



import pytest

import pyteal as pt
from pyteal.ast.comment import CommentExpr

options = pt.CompileOptions()


def test_CommentExpr():
    for comment in ("", "hello world", " // a b c //    \t . "):
        expr = CommentExpr(comment)
        assert expr.comment == comment
        assert expr.type_of() == pt.TealType.none
        assert expr.has_return() is False

        expected = pt.TealSimpleBlock(
            [
                pt.TealOp(expr, pt.Op.comment, comment),
            ]
        )

        actual, _ = expr.__teal__(options)
        actual.addIncoming()
        actual = pt.TealBlock.NormalizeBlocks(actual)

        assert actual == expected

    for newline in ("\n", "\r\n", "\r"):
        with pytest.raises(
            pt.TealInputError,
            match=r"Newlines should not be present in the CommentExpr constructor$",
        ):
            CommentExpr(f"one line{newline}two lines")


def test_Comment_Expr_empty():
    comment = "dope"
    expr = pt.Comment(comment)
    assert type(expr) is pt.Seq
    assert len(expr.args) == 1

    assert getattr(expr.args[0], "comment") == comment


def test_Comment_empty():
    to_wrap = pt.Int(1)
    comment = ""
    expr = pt.Comment(comment, to_wrap)
    assert type(expr) is pt.Seq
    assert len(expr.args) == 1

    assert expr.args[0] is to_wrap


def test_Comment_single_line():
    to_wrap = pt.Int(1)
    comment = "just an int"
    expr = pt.Comment(comment, to_wrap)
    assert type(expr) is pt.Seq
    assert len(expr.args) == 2

    assert type(expr.args[0]) is CommentExpr
    assert expr.args[0].comment == comment

    assert expr.args[1] is to_wrap

    version = 6
    expected_teal = f"""#pragma version {version}
// {comment}
int 1
return"""
    actual_teal = pt.compileTeal(
        pt.Return(expr), version=version, mode=pt.Mode.Application
    )
    assert actual_teal == expected_teal


def test_Comment_multi_line():
    to_wrap = pt.Int(1)
    comment = """just an int
but its really more than that isnt it? an integer here is a uint64 stack type but looking further what does that mean? 
You might say its a 64 bit representation of an element of the set Z and comes from the latin `integer` meaning `whole`
since it has no fractional part. You might also say this run on comment has gone too far. See https://en.wikipedia.org/wiki/Integer for more details 
"""

    comment_parts = [
        "just an int",
        "but its really more than that isnt it? an integer here is a uint64 stack type but looking further what does that mean? ",
        "You might say its a 64 bit representation of an element of the set Z and comes from the latin `integer` meaning `whole`",
        "since it has no fractional part. You might also say this run on comment has gone too far. See https://en.wikipedia.org/wiki/Integer for more details ",
    ]

    expr = pt.Comment(comment, to_wrap)
    assert type(expr) is pt.Seq
    assert len(expr.args) == 5

    for i, part in enumerate(comment_parts):
        arg = expr.args[i]
        assert type(arg) is CommentExpr
        assert arg.comment == part

    assert expr.args[4] is to_wrap

    version = 6
    expected_teal = f"""#pragma version {version}
// just an int
// but its really more than that isnt it? an integer here is a uint64 stack type but looking further what does that mean? 
// You might say its a 64 bit representation of an element of the set Z and comes from the latin `integer` meaning `whole`
// since it has no fractional part. You might also say this run on comment has gone too far. See https://en.wikipedia.org/wiki/Integer for more details 
int 1
return"""
    actual_teal = pt.compileTeal(
        pt.Return(expr), version=version, mode=pt.Mode.Application
    )

    assert actual_teal == expected_teal



#!/usr/bin/env python3

import base64
import params
import uuid

from algosdk import algod, transaction, account, mnemonic
from periodic_payment import periodic_payment

from pyteal import *

# --------- compile & send transaction using Goal and Python SDK ----------

teal_source = compileTeal(periodic_payment(), mode=Mode.Signature, version=2)

# compile teal
teal_file = str(uuid.uuid4()) + ".teal"
with open(teal_file, "w+") as f:
    f.write(teal_source)
lsig_fname = str(uuid.uuid4()) + ".tealc"

stdout, stderr = execute(["goal", "clerk", "compile", "-o", lsig_fname, teal_file])

if stderr != "":
    print(stderr)
    raise
elif len(stdout) < 59:
    print("error in compile teal")
    raise

with open(lsig_fname, "rb") as f:
    teal_bytes = f.read()
lsig = transaction.LogicSig(teal_bytes)

# create algod clients
acl = algod.AlgodClient(params.algod_token, params.algod_address)

# Recover the account that is wanting to delegate signature
passphrase = "patrol crawl rule faculty enemy sick reveal embody trumpet win shy zero ill draw swim excuse tongue under exact baby moral kite spring absent double"
sk = mnemonic.to_private_key(passphrase)
addr = account.address_from_private_key(sk)
print("Dispense at least 201000 microAlgo to {}".format(addr))
input("Make sure you did that. Press Enter to continue...")

# sign the logic signature with an account sk
lsig.sign(sk)

# get suggested parameters
params = acl.suggested_params()
gen = params["genesisID"]
gh = params["genesishashb64"]
startRound = params["lastRound"] - (params["lastRound"] % 1000)
endRound = startRound + 1000
fee = 1000
amount = 200000
receiver = "ZZAF5ARA4MEC5PVDOP64JM5O5MQST63Q2KOY2FLYFLXXD3PFSNJJBYAFZM"
lease = base64.b64decode("y9OJ5MRLCHQj8GqbikAUKMBI7hom+SOj8dlopNdNHXI=")

# create a transaction
txn = transaction.PaymentTxn(
    addr, fee, startRound, endRound, gh, receiver, amount, flat_fee=True, lease=lease
)

# Create the LogicSigTransaction with contract account LogicSig
lstx = transaction.LogicSigTransaction(txn, lsig)

# write to file
txns = [lstx]
transaction.write_to_file(txns, "p_pay.stxn")

# send raw LogicSigTransaction to network
txid = acl.send_transaction(lstx)
print("Transaction ID: " + txid)
# except Exception as e:
#     print(e)

# send raw LogicSigTransaction again to network
txid = acl.send_transaction(lstx)
print("Transaction ID: " + txid)



# This example is provided for informational purposes only and has not been audited for security.

from pyteal import *

""" Recurring Swap

This is a recurring swap contract. This contract can be set as an escrow 
and sending tmpl_amount recurringly to tmpl_provider as long as the transaction is authoried
by the provider.

Use scenarios:
A buyer (an insurer, for example) set up this escrow contract and funded with Algos. 
Whenever a service provider (a hospital, for example) authorize a payment,
it submits a transaction with its signature of first_valid in the arguement and with
the first_valid in the lease to prevent replay attacks.

The hospital may some extra information about the transaction in the note 
field for audit purpose.

After timeout, the buyer can claw the remaining balance back.
tmpl_buyer: buyer who receives the service
tmpl_provider: provider of the service
tmpl_ppk: public key of the provider
tmpl_amount: the amount of microAlgo charged per service
tmpl_fee: maximum transaction fee allowed
tmpl_timeout: the timeout round, after which the buyer can reclaim her Algos
"""

tmpl_buyer = Addr("6ZHGHH5Z5CTPCF5WCESXMGRSVK7QJETR63M3NY5FJCUYDHO57VTCMJOBGY")
tmpl_provider = Addr("7Z5PWO2C6LFNQFGHWKSK5H47IQP5OJW2M3HA2QPXTY3WTNP5NU2MHBW27M")
tmpl_amount = Int(100000)
tmpl_fee = Int(1000)
tmpl_timeout = Int(100000)


def recurring_swap(
    tmpl_buyer=tmpl_buyer,
    tmpl_provider=tmpl_provider,
    tmpl_amount=tmpl_amount,
    tmpl_fee=tmpl_fee,
    tmpl_timeout=tmpl_timeout,
):
    fee_cond = Txn.fee() <= tmpl_fee
    type_cond = And(Txn.type_enum() == Int(1), Txn.rekey_to() == Global.zero_address())
    recv_cond = And(
        Txn.close_remainder_to() == Global.zero_address(),
        Txn.receiver() == tmpl_provider,
        Txn.amount() == tmpl_amount,
        Ed25519Verify(Itob(Txn.first_valid()), Arg(0), tmpl_provider),
        Txn.lease() == Sha256(Itob(Txn.first_valid())),
    )

    close_cond = And(
        Txn.close_remainder_to() == tmpl_buyer,
        Txn.amount() == Int(0),
        Txn.first_valid() >= tmpl_timeout,
    )

    program = And(fee_cond, type_cond, Or(recv_cond, close_cond))

    return program


if __name__ == "__main__":
    print(compileTeal(recurring_swap(), mode=Mode.Signature, version=2))



from abc import ABC, abstractmethod
from typing import Tuple, List, TYPE_CHECKING

from pyteal.types import TealType
from pyteal.ir import TealBlock, TealSimpleBlock

if TYPE_CHECKING:
    from pyteal.compiler import CompileOptions


class Expr(ABC):
    """Abstract base class for PyTeal expressions."""

    def __init__(self):
        import traceback

        self.trace = traceback.format_stack()[0:-1]

    def getDefinitionTrace(self) -> List[str]:
        return self.trace

    @abstractmethod
    def type_of(self) -> TealType:
        """Get the return type of this expression."""
        pass

    @abstractmethod
    def has_return(self) -> bool:
        """Check if this expression always returns from the current subroutine or program."""
        pass

    @abstractmethod
    def __str__(self) -> str:
        """Get a string representation of this expression."""
        pass

    @abstractmethod
    def __teal__(self, options: "CompileOptions") -> Tuple[TealBlock, TealSimpleBlock]:
        """Assemble TEAL IR for this component and its arguments."""
        pass

    def __lt__(self, other):
        from pyteal.ast.binaryexpr import Lt

        return Lt(self, other)

    def __gt__(self, other):
        from pyteal.ast.binaryexpr import Gt

        return Gt(self, other)

    def __le__(self, other):
        from pyteal.ast.binaryexpr import Le

        return Le(self, other)

    def __ge__(self, other):
        from pyteal.ast.binaryexpr import Ge

        return Ge(self, other)

    def __eq__(self, other):
        from pyteal.ast.binaryexpr import Eq

        return Eq(self, other)

    def __ne__(self, other):
        from pyteal.ast.binaryexpr import Neq

        return Neq(self, other)

    def __add__(self, other):
        from pyteal.ast.naryexpr import Add

        return Add(self, other)

    def __sub__(self, other):
        from pyteal.ast.binaryexpr import Minus

        return Minus(self, other)

    def __mul__(self, other):
        from pyteal.ast.naryexpr import Mul

        return Mul(self, other)

    def __truediv__(self, other):
        from pyteal.ast.binaryexpr import Div

        return Div(self, other)

    def __mod__(self, other):
        from pyteal.ast.binaryexpr import Mod

        return Mod(self, other)

    def __pow__(self, other):
        from pyteal.ast.binaryexpr import Exp

        return Exp(self, other)

    def __invert__(self):
        from pyteal.ast.unaryexpr import BitwiseNot

        return BitwiseNot(self)

    def __and__(self, other):
        from pyteal.ast.binaryexpr import BitwiseAnd

        return BitwiseAnd(self, other)

    def __or__(self, other):
        from pyteal.ast.binaryexpr import BitwiseOr

        return BitwiseOr(self, other)

    def __xor__(self, other):
        from pyteal.ast.binaryexpr import BitwiseXor

        return BitwiseXor(self, other)

    def __lshift__(self, other):
        from pyteal.ast.binaryexpr import ShiftLeft

        return ShiftLeft(self, other)

    def __rshift__(self, other):
        from pyteal.ast.binaryexpr import ShiftRight

        return ShiftRight(self, other)

    def And(self, other: "Expr") -> "Expr":
        """Take the logical And of this expression and another one.

        This expression must evaluate to uint64.

        This is the same as using :func:`And()` with two arguments.
        """
        from pyteal.ast.naryexpr import And

        return And(self, other)

    def Or(self, other: "Expr") -> "Expr":
        """Take the logical Or of this expression and another one.

        This expression must evaluate to uint64.

        This is the same as using :func:`Or()` with two arguments.
        """
        from pyteal.ast.naryexpr import Or

        return Or(self, other)


Expr.__module__ = "pyteal"



import pytest

import pyteal as pt

avm2Options = pt.CompileOptions(version=2)
avm3Options = pt.CompileOptions(version=3)
avm4Options = pt.CompileOptions(version=4)
avm5Options = pt.CompileOptions(version=5)


def test_substring_immediate_v2():
    args = [pt.Bytes("my string"), pt.Int(0), pt.Int(2)]
    expr = pt.Substring(args[0], args[1], args[2])
    assert expr.type_of() == pt.TealType.bytes

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.byte, '"my string"'),
            pt.TealOp(expr, pt.Op.substring, 0, 2),
        ]
    )

    actual, _ = expr.__teal__(avm2Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected


def test_substring_immediate_v5():
    args = [pt.Bytes("my string"), pt.Int(1), pt.Int(2)]
    expr = pt.Substring(args[0], args[1], args[2])
    assert expr.type_of() == pt.TealType.bytes

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.byte, '"my string"'),
            pt.TealOp(expr, pt.Op.extract, 1, 1),
        ]
    )

    actual, _ = expr.__teal__(avm5Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected


def test_substring_to_extract():
    my_string = "a" * 257
    args = [pt.Bytes(my_string), pt.Int(255), pt.Int(257)]
    expr = pt.Substring(args[0], args[1], args[2])
    assert expr.type_of() == pt.TealType.bytes

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.byte, '"{my_string}"'.format(my_string=my_string)),
            pt.TealOp(expr, pt.Op.extract, 255, 2),
        ]
    )

    actual, _ = expr.__teal__(avm5Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def test_substring_stack_v2():
    my_string = "a" * 257
    args = [pt.Bytes(my_string), pt.Int(256), pt.Int(257)]
    expr = pt.Substring(args[0], args[1], args[2])
    assert expr.type_of() == pt.TealType.bytes

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.byte, '"{my_string}"'.format(my_string=my_string)),
            pt.TealOp(args[1], pt.Op.int, 256),
            pt.TealOp(args[2], pt.Op.int, 257),
            pt.TealOp(expr, pt.Op.substring3),
        ]
    )

    actual, _ = expr.__teal__(avm2Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def test_substring_stack_v5():
    my_string = "a" * 257
    args = [pt.Bytes(my_string), pt.Int(256), pt.Int(257)]
    expr = pt.Substring(args[0], args[1], args[2])
    assert expr.type_of() == pt.TealType.bytes

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.byte, '"{my_string}"'.format(my_string=my_string)),
            pt.TealOp(args[1], pt.Op.int, 256),
            pt.TealOp(pt.Int(1), pt.Op.int, 1),
            pt.TealOp(expr, pt.Op.extract3),
        ]
    )

    actual, _ = expr.__teal__(avm5Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def test_zero_length_substring_immediate():
    my_string = "a" * 257
    args = [pt.Bytes(my_string), pt.Int(1), pt.Int(1)]
    expr = pt.Substring(args[0], args[1], args[2])
    assert expr.type_of() == pt.TealType.bytes

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.byte, '"{my_string}"'.format(my_string=my_string)),
            pt.TealOp(expr, pt.Op.substring, 1, 1),
        ]
    )

    actual, _ = expr.__teal__(avm5Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def test_substring_invalid():
    with pytest.raises(pt.TealTypeError):
        pt.Substring(pt.Int(0), pt.Int(0), pt.Int(2))

    with pytest.raises(pt.TealTypeError):
        pt.Substring(pt.Bytes("my string"), pt.Txn.sender(), pt.Int(2))

    with pytest.raises(pt.TealTypeError):
        pt.Substring(pt.Bytes("my string"), pt.Int(0), pt.Txn.sender())

    with pytest.raises(Exception):
        pt.Substring(pt.Bytes("my string"), pt.Int(1), pt.Int(0)).__teal__(avm5Options)


def test_extract_immediate():
    args = [pt.Bytes("my string"), pt.Int(0), pt.Int(2)]
    expr = pt.Extract(args[0], args[1], args[2])
    assert expr.type_of() == pt.TealType.bytes

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.byte, '"my string"'),
            pt.TealOp(expr, pt.Op.extract, 0, 2),
        ]
    )

    actual, _ = expr.__teal__(avm5Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected

    with pytest.raises(pt.TealInputError):
        expr.__teal__(avm4Options)


def test_extract_zero():
    args = [pt.Bytes("my string"), pt.Int(1), pt.Int(0)]
    expr = pt.Extract(args[0], args[1], args[2])
    assert expr.type_of() == pt.TealType.bytes

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.byte, '"my string"'),
            pt.TealOp(args[1], pt.Op.int, 1),
            pt.TealOp(args[2], pt.Op.int, 0),
            pt.TealOp(expr, pt.Op.extract3),
        ]
    )

    actual, _ = expr.__teal__(avm5Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected

    with pytest.raises(pt.TealInputError):
        expr.__teal__(avm4Options)


def test_extract_stack():
    my_string = "*" * 257
    args = [pt.Bytes(my_string), pt.Int(256), pt.Int(257)]
    expr = pt.Extract(args[0], args[1], args[2])
    assert expr.type_of() == pt.TealType.bytes

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.byte, '"{my_string}"'.format(my_string=my_string)),
            pt.TealOp(args[1], pt.Op.int, 256),
            pt.TealOp(args[2], pt.Op.int, 257),
            pt.TealOp(expr, pt.Op.extract3),
        ]
    )

    actual, _ = expr.__teal__(avm5Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected

    with pytest.raises(pt.TealInputError):
        expr.__teal__(avm4Options)


def test_extract_invalid():
    with pytest.raises(pt.TealTypeError):
        pt.Extract(pt.Int(0), pt.Int(0), pt.Int(2))

    with pytest.raises(pt.TealTypeError):
        pt.Extract(pt.Bytes("my string"), pt.Txn.sender(), pt.Int(2))

    with pytest.raises(pt.TealTypeError):
        pt.Extract(pt.Bytes("my string"), pt.Int(0), pt.Txn.sender())


def test_suffix_immediate():
    args = [pt.Bytes("my string"), pt.Int(1)]
    expr = pt.Suffix(args[0], args[1])
    assert expr.type_of() == pt.TealType.bytes

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.byte, '"my string"'),
            pt.TealOp(expr, pt.Op.extract, 1, 0),
        ]
    )

    actual, _ = expr.__teal__(avm5Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected


def test_suffix_stack():
    my_string = "*" * 1024
    args = [pt.Bytes(my_string), pt.Int(257)]
    expr = pt.Suffix(args[0], args[1])
    assert expr.type_of() == pt.TealType.bytes

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.byte, '"{my_string}"'.format(my_string=my_string)),
            pt.TealOp(args[1], pt.Op.int, 257),
            pt.TealOp(expr, pt.Op.dig, 1),
            pt.TealOp(expr, pt.Op.len),
            pt.TealOp(expr, pt.Op.substring3),
        ]
    )

    actual, _ = expr.__teal__(avm2Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected


@pytest.mark.parametrize("op", [pt.Op.extract3, pt.Op.substring3])
def test_startArg_not_int(op: pt.Op):
    my_string = "*" * 257
    add = pt.Add(pt.Int(254), pt.Int(2))
    args = [pt.Bytes(my_string), add, pt.Int(257)]

    def generate_expr() -> pt.Expr:
        match op:
            case pt.Op.extract3:
                return pt.Extract(args[0], args[1], args[2])
            case pt.Op.substring3:
                return pt.Substring(args[0], args[1], args[2])
            case _:
                raise Exception(f"Unsupported {op=}")

    expr = generate_expr()
    assert expr.type_of() == pt.TealType.bytes

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.byte, '"{my_string}"'.format(my_string=my_string)),
            pt.TealOp(pt.Int(254), pt.Op.int, 254),
            pt.TealOp(pt.Int(2), pt.Op.int, 2),
            pt.TealOp(add, pt.Op.add),
            pt.TealOp(args[2], pt.Op.int, 257),
            pt.TealOp(None, op),
        ]
    )

    actual, _ = expr.__teal__(avm5Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected

    if op == pt.Op.extract3:
        with pytest.raises(pt.TealInputError):
            expr.__teal__(avm4Options)


@pytest.mark.parametrize("op", [pt.Op.extract3, pt.Op.substring3])
def test_endArg_not_int(op: pt.Op):
    my_string = "*" * 257
    add = pt.Add(pt.Int(254), pt.Int(3))
    args = [pt.Bytes(my_string), pt.Int(256), add]

    def generate_expr() -> pt.Expr:
        match op:
            case pt.Op.extract3:
                return pt.Extract(args[0], args[1], args[2])
            case pt.Op.substring3:
                return pt.Substring(args[0], args[1], args[2])
            case _:
                raise Exception(f"Unsupported {op=}")

    expr = generate_expr()
    assert expr.type_of() == pt.TealType.bytes

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.byte, '"{my_string}"'.format(my_string=my_string)),
            pt.TealOp(args[1], pt.Op.int, 256),
            pt.TealOp(pt.Int(254), pt.Op.int, 254),
            pt.TealOp(pt.Int(3), pt.Op.int, 3),
            pt.TealOp(add, pt.Op.add),
            pt.TealOp(None, op),
        ]
    )

    actual, _ = expr.__teal__(avm5Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected

    if op == pt.Op.extract3:
        with pytest.raises(pt.TealInputError):
            expr.__teal__(avm4Options)


def test_suffix_invalid():
    with pytest.raises(pt.TealTypeError):
        pt.Suffix(pt.Int(0), pt.Int(0))

    with pytest.raises(pt.TealTypeError):
        pt.Suffix(pt.Bytes("my string"), pt.Txn.sender())



import base64

from typing import Union, List, Dict, cast
from collections import OrderedDict
from algosdk import encoding

from pyteal.ir import (
    Op,
    TealOp,
    TealComponent,
)
from pyteal.util import unescapeStr, correctBase32Padding
from pyteal.errors import TealInternalError

intEnumValues = {
    # OnComplete values
    "NoOp": 0,
    "OptIn": 1,
    "CloseOut": 2,
    "ClearState": 3,
    "UpdateApplication": 4,
    "DeleteApplication": 5,
    # TxnType values
    "unknown": 0,
    "pay": 1,
    "keyreg": 2,
    "acfg": 3,
    "axfer": 4,
    "afrz": 5,
    "appl": 6,
}


def extractIntValue(op: TealOp) -> Union[str, int]:
    """Extract the constant value being loaded by a TealOp whose op is Op.int.

    Returns:
        If the op is loading a template variable, returns the name of the variable as a string.
        Otherwise, returns the integer that the op is loading.
    """
    if len(op.args) != 1 or type(op.args[0]) not in (int, str):
        raise TealInternalError("Unexpected args in int opcode: {}".format(op.args))

    value = cast(Union[str, int], op.args[0])
    if type(value) is int or cast(str, value).startswith("TMPL_"):
        return value
    if value not in intEnumValues:
        raise TealInternalError("Int constant not recognized: {}".format(value))
    return intEnumValues[cast(str, value)]


def extractBytesValue(op: TealOp) -> Union[str, bytes]:
    """Extract the constant value being loaded by a TealOp whose op is Op.byte.

    Returns:
        If the op is loading a template variable, returns the name of the variable as a string.
        Otherwise, returns the byte string that the op is loading.
    """
    if len(op.args) != 1 or type(op.args[0]) is not str:
        raise TealInternalError("Unexpected args in byte opcode: {}".format(op.args))

    value = op.args[0]
    if value.startswith("TMPL_"):
        return value
    if value.startswith('"') and value.endswith('"'):
        return unescapeStr(value).encode("utf-8")
    if value.startswith("0x"):
        return bytes.fromhex(value[2:])
    if value.startswith("base32(") and value.endswith(")"):
        return base64.b32decode(correctBase32Padding(value[len("base32(") : -1]))
    if value.startswith("base64(") and value.endswith(")"):
        return base64.b64decode(value[len("base64(") : -1])

    raise TealInternalError("Unexpected format for byte value: {}".format(value))


def extractAddrValue(op: TealOp) -> Union[str, bytes]:
    """Extract the constant value being loaded by a TealOp whose op is Op.addr.

    Returns:
        If the op is loading a template variable, returns the name of the variable as a string.
        Otherwise, returns the bytes of the public key of the address that the op is loading.
    """
    if len(op.args) != 1 or type(op.args[0]) != str:
        raise TealInternalError("Unexpected args in addr opcode: {}".format(op.args))

    value = cast(str, op.args[0])
    if not value.startswith("TMPL_"):
        value = encoding.decode_address(value)
    return value


def extractMethodSigValue(op: TealOp) -> bytes:
    """Extract the constant value being loaded by a TealOp whose op is Op.method.

    Returns:
        The bytes of method selector computed from the method signature that the op is loading.
    """
    if len(op.args) != 1 or type(op.args[0]) != str:
        raise TealInternalError("Unexpected args in method opcode: {}".format(op.args))

    methodSignature = cast(str, op.args[0])
    if methodSignature[0] == methodSignature[-1] and methodSignature.startswith('"'):
        methodSignature = methodSignature[1:-1]
    else:
        raise TealInternalError(
            "Method signature opcode error: signature {} not wrapped with double-quotes".format(
                methodSignature
            )
        )
    methodSelector = encoding.checksum(bytes(methodSignature, "utf-8"))[:4]
    return methodSelector


def createConstantBlocks(ops: List[TealComponent]) -> List[TealComponent]:
    """Convert TEAL code from using pseudo-ops for constants to using assembled constant blocks.

    This conversion will assemble constants to be as space-efficient as possible.

    Args:
        ops: A list of TealComponents to convert.

    Returns:
        A list of TealComponent that are functionally the same as the input, but with all constants
        loaded either through blocks or the `pushint`/`pushbytes` single-use ops.
    """
    intFreqs: Dict[Union[str, int], int] = OrderedDict()
    byteFreqs: Dict[Union[str, bytes], int] = OrderedDict()

    for op in ops:
        if not isinstance(op, TealOp):
            continue

        basicOp = op.getOp()

        if basicOp == Op.int:
            intValue = extractIntValue(op)
            intFreqs[intValue] = intFreqs.get(intValue, 0) + 1
        elif basicOp == Op.byte:
            byteValue = extractBytesValue(op)
            byteFreqs[byteValue] = byteFreqs.get(byteValue, 0) + 1
        elif basicOp == Op.addr:
            addrValue = extractAddrValue(op)
            byteFreqs[addrValue] = byteFreqs.get(addrValue, 0) + 1
        elif basicOp == Op.method_signature:
            methodValue = extractMethodSigValue(op)
            byteFreqs[methodValue] = byteFreqs.get(methodValue, 0) + 1

    assembled: List[TealComponent] = []

    # because we used OrderedDicts and python sorting is stable, constants with the same frequency
    # will remain in the same order, i.e. first defined, first in block
    sortedInts = sorted(intFreqs, key=lambda x: intFreqs[x], reverse=True)
    sortedBytes = sorted(byteFreqs, key=lambda x: byteFreqs[x], reverse=True)

    # Use Op.pushint if the constant does not occur in the top 4 most frequent and is smaller than
    # 2 ** 7 to improve performance and save block space.
    intBlock = [
        val
        for i, val in enumerate(sortedInts)
        if intFreqs[val] > 1 and (i < 4 or isinstance(val, str) or val >= 2**7)
    ]

    byteBlock = [
        ("0x" + b.hex()) if type(b) is bytes else cast(str, b)
        for b in sortedBytes
        if byteFreqs[b] > 1
    ]

    if len(intBlock) != 0:
        assembled.append(TealOp(None, Op.intcblock, *intBlock))

    if len(byteBlock) != 0:
        assembled.append(TealOp(None, Op.bytecblock, *byteBlock))

    for op in ops:
        if isinstance(op, TealOp):
            basicOp = op.getOp()

            if basicOp == Op.int:
                intValue = extractIntValue(op)
                if intValue not in intBlock:
                    assembled.append(
                        TealOp(op.expr, Op.pushint, intValue, "//", *op.args)
                    )
                    continue

                index = intBlock.index(intValue)
                if index == 0:
                    assembled.append(TealOp(op.expr, Op.intc_0, "//", *op.args))
                elif index == 1:
                    assembled.append(TealOp(op.expr, Op.intc_1, "//", *op.args))
                elif index == 2:
                    assembled.append(TealOp(op.expr, Op.intc_2, "//", *op.args))
                elif index == 3:
                    assembled.append(TealOp(op.expr, Op.intc_3, "//", *op.args))
                else:
                    assembled.append(TealOp(op.expr, Op.intc, index, "//", *op.args))
                continue

            if (
                basicOp == Op.byte
                or basicOp == Op.addr
                or basicOp == Op.method_signature
            ):
                if basicOp == Op.byte:
                    byteValue = extractBytesValue(op)
                elif basicOp == Op.addr:
                    byteValue = extractAddrValue(op)
                elif basicOp == Op.method_signature:
                    byteValue = extractMethodSigValue(op)
                else:
                    raise TealInternalError(
                        "Expect a byte-like constant opcode, get {}".format(op)
                    )

                if byteFreqs[byteValue] == 1:
                    encodedValue = (
                        ("0x" + byteValue.hex())
                        if type(byteValue) is bytes
                        else cast(str, byteValue)
                    )
                    assembled.append(
                        TealOp(op.expr, Op.pushbytes, encodedValue, "//", *op.args)
                    )
                    continue

                index = sortedBytes.index(byteValue)
                if index == 0:
                    assembled.append(TealOp(op.expr, Op.bytec_0, "//", *op.args))
                elif index == 1:
                    assembled.append(TealOp(op.expr, Op.bytec_1, "//", *op.args))
                elif index == 2:
                    assembled.append(TealOp(op.expr, Op.bytec_2, "//", *op.args))
                elif index == 3:
                    assembled.append(TealOp(op.expr, Op.bytec_3, "//", *op.args))
                else:
                    assembled.append(TealOp(op.expr, Op.bytec, index, "//", *op.args))
                continue

        assembled.append(op)

    return assembled



from typing import TYPE_CHECKING

from pyteal.types import TealType
from pyteal.ir import TealOp, Op, TealBlock
from pyteal.errors import TealInputError
from pyteal.ast.leafexpr import LeafExpr

if TYPE_CHECKING:
    from pyteal.compiler import CompileOptions


class Int(LeafExpr):
    """An expression that represents a uint64."""

    def __init__(self, value: int) -> None:
        """Create a new uint64.

        Args:
            value: The integer value this uint64 will represent. Must be a positive value less than
                2**64.
        """
        super().__init__()

        if type(value) is not int:
            raise TealInputError("invalid input type {} to Int".format(type(value)))
        elif value >= 0 and value < 2**64:
            self.value = value
        else:
            raise TealInputError("Int {} is out of range".format(value))

    def __teal__(self, options: "CompileOptions"):
        op = TealOp(self, Op.int, self.value)
        return TealBlock.FromOp(options, op)

    def __str__(self):
        return "(Int {})".format(self.value)

    def type_of(self):
        return TealType.uint64


Int.__module__ = "pyteal"


class EnumInt(LeafExpr):
    """An expression that represents uint64 enum values."""

    def __init__(self, name: str) -> None:
        """Create an expression to reference a uint64 enum value.

        Args:
            name: The name of the enum value.
        """
        super().__init__()
        self.name = name

    def __teal__(self, options: "CompileOptions"):
        op = TealOp(self, Op.int, self.name)
        return TealBlock.FromOp(options, op)

    def __str__(self):
        return "(IntEnum {})".format(self.name)

    def type_of(self):
        return TealType.uint64


EnumInt.__module__ = "pyteal"



from typing import cast, TYPE_CHECKING

from pyteal.types import TealType, require_type
from pyteal.errors import TealCompileError, verifyProgramVersion
from pyteal.ir import TealOp, Op, TealBlock, TealSimpleBlock
from pyteal.ast.expr import Expr
from pyteal.ast.int import Int
from pyteal.ast.ternaryexpr import TernaryExpr

if TYPE_CHECKING:
    from pyteal.compiler import CompileOptions


class SubstringExpr(Expr):
    """An expression for taking the substring of a byte string given start and end indices"""

    def __init__(self, stringArg: Expr, startArg: Expr, endArg: Expr) -> None:
        super().__init__()

        require_type(stringArg, TealType.bytes)
        require_type(startArg, TealType.uint64)
        require_type(endArg, TealType.uint64)

        self.stringArg = stringArg
        self.startArg = startArg
        self.endArg = endArg

    # helper method for correctly populating op
    def __get_op(self, options: "CompileOptions"):
        s, e = cast(Int, self.startArg).value, cast(Int, self.endArg).value
        l = e - s

        if l < 0:
            raise TealCompileError(
                "The end index must be greater than or equal to the start index",
                self,
            )

        if l > 0 and options.version >= Op.extract.min_version:
            if s < 2**8 and l < 2**8:
                return Op.extract
            else:
                return Op.extract3
        else:
            if s < 2**8 and e < 2**8:
                return Op.substring
            else:
                return Op.substring3

    def __teal__(self, options: "CompileOptions"):
        if not isinstance(self.startArg, Int) or not isinstance(self.endArg, Int):
            return TernaryExpr(
                Op.substring3,
                (TealType.bytes, TealType.uint64, TealType.uint64),
                TealType.bytes,
                self.stringArg,
                self.startArg,
                self.endArg,
            ).__teal__(options)

        op = self.__get_op(options)

        verifyProgramVersion(
            op.min_version,
            options.version,
            "Program version too low to use op {}".format(op),
        )

        start, end = cast(Int, self.startArg).value, cast(Int, self.endArg).value
        if op == Op.extract:
            length = end - start
            return TealBlock.FromOp(
                options,
                TealOp(self, op, self.startArg.value, length),
                self.stringArg,
            )
        elif op == Op.extract3:
            length = end - start
            return TealBlock.FromOp(
                options,
                TealOp(self, op),
                self.stringArg,
                self.startArg,
                Int(length),
            )
        elif op == Op.substring:
            return TealBlock.FromOp(
                options, TealOp(self, op, start, end), self.stringArg
            )
        elif op == Op.substring3:
            return TealBlock.FromOp(
                options,
                TealOp(self, op),
                self.stringArg,
                self.startArg,
                self.endArg,
            )

    def __str__(self):
        return "(Substring {} {} {})".format(self.stringArg, self.startArg, self.endArg)

    def type_of(self):
        return TealType.bytes

    def has_return(self):
        return False


class ExtractExpr(Expr):
    """An expression for extracting a section of a byte string given a start index and length"""

    def __init__(self, stringArg: Expr, startArg: Expr, lenArg: Expr) -> None:
        super().__init__()

        require_type(stringArg, TealType.bytes)
        require_type(startArg, TealType.uint64)
        require_type(lenArg, TealType.uint64)

        self.stringArg = stringArg
        self.startArg = startArg
        self.lenArg = lenArg

    # helper method for correctly populating op
    def __get_op(self, options: "CompileOptions"):
        s, l = cast(Int, self.startArg).value, cast(Int, self.lenArg).value
        if s < 2**8 and l > 0 and l < 2**8:
            return Op.extract
        else:
            return Op.extract3

    def __teal__(self, options: "CompileOptions"):
        if not isinstance(self.startArg, Int) or not isinstance(self.lenArg, Int):
            return TernaryExpr(
                Op.extract3,
                (TealType.bytes, TealType.uint64, TealType.uint64),
                TealType.bytes,
                self.stringArg,
                self.startArg,
                self.lenArg,
            ).__teal__(options)

        op = self.__get_op(options)

        verifyProgramVersion(
            op.min_version,
            options.version,
            "Program version too low to use op {}".format(op),
        )

        s, l = cast(Int, self.startArg).value, cast(Int, self.lenArg).value
        if op == Op.extract:
            return TealBlock.FromOp(options, TealOp(self, op, s, l), self.stringArg)
        elif op == Op.extract3:
            return TealBlock.FromOp(
                options,
                TealOp(self, op),
                self.stringArg,
                self.startArg,
                self.lenArg,
            )

    def __str__(self):
        return "(Extract {} {} {})".format(self.stringArg, self.startArg, self.lenArg)

    def type_of(self):
        return TealType.bytes

    def has_return(self):
        return False


class SuffixExpr(Expr):
    """An expression for taking the suffix of a byte string given start index"""

    def __init__(
        self,
        stringArg: Expr,
        startArg: Expr,
    ) -> None:
        super().__init__()

        require_type(stringArg, TealType.bytes)
        require_type(startArg, TealType.uint64)

        self.stringArg = stringArg
        self.startArg = startArg

    # helper method for correctly populating op
    def __get_op(self, options: "CompileOptions"):
        if not isinstance(self.startArg, Int):
            return Op.substring3

        s = cast(Int, self.startArg).value
        if s < 2**8:
            return Op.extract
        else:
            return Op.substring3

    def __teal__(self, options: "CompileOptions"):
        op = self.__get_op(options)

        verifyProgramVersion(
            op.min_version,
            options.version,
            "Program version too low to use op {}".format(op),
        )

        if op == Op.extract:
            # if possible, exploit optimization in the extract opcode that takes the suffix
            # when the length argument is 0
            return TealBlock.FromOp(
                options,
                TealOp(self, op, cast(Int, self.startArg).value, 0),
                self.stringArg,
            )
        elif op == Op.substring3:
            strBlockStart, strBlockEnd = self.stringArg.__teal__(options)
            nextBlockStart, nextBlockEnd = self.startArg.__teal__(options)
            strBlockEnd.setNextBlock(nextBlockStart)

            finalBlock = TealSimpleBlock(
                [
                    TealOp(self, Op.dig, 1),
                    TealOp(self, Op.len),
                    TealOp(self, Op.substring3),
                ]
            )

            nextBlockEnd.setNextBlock(finalBlock)
            return strBlockStart, finalBlock

    def __str__(self):
        return "(Suffix {} {})".format(self.stringArg, self.startArg)

    def type_of(self):
        return TealType.bytes

    def has_return(self):
        return False


def Substring(string: Expr, start: Expr, end: Expr) -> Expr:
    """Take a substring of a byte string.

    Produces a new byte string consisting of the bytes starting at :code:`start` up to but not
    including :code:`end`.

    This expression is similar to :any:`Extract`, except this expression uses start and end indexes,
    while :code:`Extract` uses a start index and length.

    Requires program version 2 or higher.

    Args:
        string: The byte string.
        start: The starting index for the substring. Must be an integer less than or equal to
            :code:`Len(string)`.
        end: The ending index for the substring. Must be an integer greater or equal to start, but
            less than or equal to Len(string).
    """
    return SubstringExpr(
        string,
        start,
        end,
    )


def Extract(string: Expr, start: Expr, length: Expr) -> Expr:
    """Extract a section of a byte string.

    Produces a new byte string consisting of the bytes starting at :code:`start` up to but not
    including :code:`start + length`.

    This expression is similar to :any:`Substring`, except this expression uses a start index and
    length, while :code:`Substring` uses start and end indexes.

    Requires program version 5 or higher.

    Args:
        string: The byte string.
        start: The starting index for the extraction. Must be an integer less than or equal to
            :code:`Len(string)`.
        length: The number of bytes to extract. Must be an integer such that :code:`start + length <= Len(string)`.
    """
    return ExtractExpr(
        string,
        start,
        length,
    )


def Suffix(string: Expr, start: Expr) -> Expr:
    """Take a suffix of a byte string.

    Produces a new byte string consisting of the suffix of the byte string starting at :code:`start`

    This expression is similar to :any:`Substring` and :any:`Extract`, except this expression only uses a
    start index.

    Requires program version 5 or higher.

    Args:
        string: The byte string.
        start: The starting index for the suffix. Must be an integer less than or equal to :code:`Len(string)`.
    """
    return SuffixExpr(
        string,
        start,
    )



import pytest

import pyteal as pt

options = pt.CompileOptions()


def test_break_fail():

    with pytest.raises(pt.TealCompileError):
        pt.Break().__teal__(options)

    with pytest.raises(pt.TealCompileError):
        pt.If(pt.Int(1), pt.Break()).__teal__(options)

    with pytest.raises(pt.TealCompileError):
        pt.Seq([pt.Break()]).__teal__(options)

    with pytest.raises(TypeError):
        pt.Break(pt.Int(1))


def test_break():

    expr = pt.Break()

    assert expr.type_of() == pt.TealType.none
    assert not expr.has_return()

    expected = pt.TealSimpleBlock([])

    options.enterLoop()
    actual, _ = expr.__teal__(options)
    breakBlocks, continueBlocks = options.exitLoop()

    assert actual == expected
    assert breakBlocks == [actual]
    assert continueBlocks == []



from typing import TYPE_CHECKING, Final
from enum import Enum

from pyteal.types import TealType, require_type
from pyteal.ir import TealOp, Op, TealBlock
from pyteal.ast.leafexpr import LeafExpr
from pyteal.ast.expr import Expr
from pyteal.ast.maybe import MaybeValue
from pyteal.ast.int import EnumInt
from pyteal.ast.global_ import Global

if TYPE_CHECKING:
    from pyteal.compiler import CompileOptions


class OnComplete:
    """An enum of values that :any:`TxnObject.on_completion()` may return."""

    NoOp = EnumInt("NoOp")
    OptIn = EnumInt("OptIn")
    CloseOut = EnumInt("CloseOut")
    ClearState = EnumInt("ClearState")
    UpdateApplication = EnumInt("UpdateApplication")
    DeleteApplication = EnumInt("DeleteApplication")


OnComplete.__module__ = "pyteal"


class AppField(Enum):
    """Enum of app fields used to create :any:`App` objects."""

    optedIn = (Op.app_opted_in, TealType.uint64)
    localGet = (Op.app_local_get, TealType.anytype)
    localGetEx = (Op.app_local_get_ex, TealType.none)
    globalGet = (Op.app_global_get, TealType.anytype)
    globalGetEx = (Op.app_global_get_ex, TealType.none)
    localPut = (Op.app_local_put, TealType.none)
    globalPut = (Op.app_global_put, TealType.none)
    localDel = (Op.app_local_del, TealType.none)
    globalDel = (Op.app_global_del, TealType.none)

    def __init__(self, op: Op, type: TealType) -> None:
        self.op = op
        self.ret_type = type

    def get_op(self) -> Op:
        return self.op

    def type_of(self) -> TealType:
        return self.ret_type


AppField.__module__ = "pyteal"


class App(LeafExpr):
    """An expression related to applications."""

    def __init__(self, field: AppField, args) -> None:
        super().__init__()
        self.field = field
        self.args = args

    def __str__(self):
        ret_str = "({}".format(self.field.get_op())
        for a in self.args:
            ret_str += " " + a.__str__()
        ret_str += ")"
        return ret_str

    def __teal__(self, options: "CompileOptions"):
        return TealBlock.FromOp(options, TealOp(self, self.field.get_op()), *self.args)

    def type_of(self):
        return self.field.type_of()

    @classmethod
    def id(cls) -> Global:
        """Get the ID of the current running application.

        This is the same as :any:`Global.current_application_id()`.
        """
        return Global.current_application_id()

    @classmethod
    def optedIn(cls, account: Expr, app: Expr) -> "App":
        """Check if an account has opted in for an application.

        Args:
            account: An index into Txn.Accounts that corresponds to the account to check,
                must be evaluated to uint64 (or, since v4, an account address that appears in
                Txn.Accounts or is Txn.Sender, must be evaluated to bytes).
            app: An index into Txn.applications that corresponds to the application to read from,
                must be evaluated to uint64 (or, since v4, an application id that appears in
                Txn.applications or is the CurrentApplicationID, must be evaluated to int).
        """
        require_type(account, TealType.anytype)
        require_type(app, TealType.uint64)
        return cls(AppField.optedIn, [account, app])

    @classmethod
    def localGet(cls, account: Expr, key: Expr) -> "App":
        """Read from an account's local state for the current application.

        Args:
            account: An index into Txn.Accounts that corresponds to the account to check,
                must be evaluated to uint64 (or, since v4, an account address that appears in
                Txn.Accounts or is Txn.Sender, must be evaluated to bytes).
            key: The key to read from the account's local state. Must evaluate to bytes.
        """
        require_type(account, TealType.anytype)
        require_type(key, TealType.bytes)
        return cls(AppField.localGet, [account, key])

    @classmethod
    def localGetEx(cls, account: Expr, app: Expr, key: Expr) -> MaybeValue:
        """Read from an account's local state for an application.

        Args:
            account: An index into Txn.Accounts that corresponds to the account to check,
                must be evaluated to uint64 (or, since v4, an account address that appears in
                Txn.Accounts or is Txn.Sender, must be evaluated to bytes).
            app: An index into Txn.applications that corresponds to the application to read from,
                must be evaluated to uint64 (or, since v4, an application id that appears in
                Txn.applications or is the CurrentApplicationID, must be evaluated to int).
            key: The key to read from the account's local state. Must evaluate to bytes.
        """
        require_type(account, TealType.anytype)
        require_type(app, TealType.uint64)
        require_type(key, TealType.bytes)
        return MaybeValue(
            AppField.localGetEx.get_op(), TealType.anytype, args=[account, app, key]
        )

    @classmethod
    def globalGet(cls, key: Expr) -> "App":
        """Read from the global state of the current application.

        Args:
            key: The key to read from the global application state. Must evaluate to bytes.
        """
        require_type(key, TealType.bytes)
        return cls(AppField.globalGet, [key])

    @classmethod
    def globalGetEx(cls, app: Expr, key: Expr) -> MaybeValue:
        """Read from the global state of an application.

        Args:
            app: An index into Txn.applications that corresponds to the application to read from,
                must be evaluated to uint64 (or, since v4, an application id that appears in
                Txn.applications or is the CurrentApplicationID, must be evaluated to uint64).
            key: The key to read from the global application state. Must evaluate to bytes.
        """
        require_type(app, TealType.uint64)
        require_type(key, TealType.bytes)
        return MaybeValue(
            AppField.globalGetEx.get_op(), TealType.anytype, args=[app, key]
        )

    @classmethod
    def localPut(cls, account: Expr, key: Expr, value: Expr) -> "App":
        """Write to an account's local state for the current application.

        Args:
            account: An index into Txn.Accounts that corresponds to the account to check,
                must be evaluated to uint64 (or, since v4, an account address that appears in
                Txn.Accounts or is Txn.Sender, must be evaluated to bytes).
            key: The key to write in the account's local state. Must evaluate to bytes.
            value: The value to write in the account's local state. Can evaluate to any type.
        """
        require_type(account, TealType.anytype)
        require_type(key, TealType.bytes)
        require_type(value, TealType.anytype)
        return cls(AppField.localPut, [account, key, value])

    @classmethod
    def globalPut(cls, key: Expr, value: Expr) -> "App":
        """Write to the global state of the current application.

        Args:
            key: The key to write in the global application state. Must evaluate to bytes.
            value: THe value to write in the global application state. Can evaluate to any type.
        """
        require_type(key, TealType.bytes)
        require_type(value, TealType.anytype)
        return cls(AppField.globalPut, [key, value])

    @classmethod
    def localDel(cls, account: Expr, key: Expr) -> "App":
        """Delete a key from an account's local state for the current application.

        Args:
            account: An index into Txn.Accounts that corresponds to the account to check,
                must be evaluated to uint64 (or, since v4, an account address that appears in
                Txn.Accounts or is Txn.Sender, must be evaluated to bytes).
            key: The key to delete from the account's local state. Must evaluate to bytes.
        """
        require_type(account, TealType.anytype)
        require_type(key, TealType.bytes)
        return cls(AppField.localDel, [account, key])

    @classmethod
    def globalDel(cls, key: Expr) -> "App":
        """Delete a key from the global state of the current application.

        Args:
            key: The key to delete from the global application state. Must evaluate to bytes.
        """
        require_type(key, TealType.bytes)
        return cls(AppField.globalDel, [key])


App.__module__ = "pyteal"


class AppParam:
    @classmethod
    def approvalProgram(cls, app: Expr) -> MaybeValue:
        """Get the bytecode of Approval Program for the application.

        Args:
            app: An index into Txn.applications that correspond to the application to check.
                Must evaluate to uint64.
        """
        require_type(app, TealType.uint64)
        return MaybeValue(
            Op.app_params_get,
            TealType.bytes,
            immediate_args=["AppApprovalProgram"],
            args=[app],
        )

    @classmethod
    def clearStateProgram(cls, app: Expr) -> MaybeValue:
        """Get the bytecode of Clear State Program for the application.

        Args:
            app: An index into Txn.applications that correspond to the application to check.
                Must evaluate to uint64.
        """
        require_type(app, TealType.uint64)
        return MaybeValue(
            Op.app_params_get,
            TealType.bytes,
            immediate_args=["AppClearStateProgram"],
            args=[app],
        )

    @classmethod
    def globalNumUint(cls, app: Expr) -> MaybeValue:
        """Get the number of uint64 values allowed in Global State for the application.

        Args:
            app: An index into Txn.applications that correspond to the application to check.
                Must evaluate to uint64.
        """
        require_type(app, TealType.uint64)
        return MaybeValue(
            Op.app_params_get,
            TealType.uint64,
            immediate_args=["AppGlobalNumUint"],
            args=[app],
        )

    @classmethod
    def globalNumByteSlice(cls, app: Expr) -> MaybeValue:
        """Get the number of byte array values allowed in Global State for the application.

        Args:
            app: An index into Txn.applications that correspond to the application to check.
                Must evaluate to uint64.
        """
        require_type(app, TealType.uint64)
        return MaybeValue(
            Op.app_params_get,
            TealType.uint64,
            immediate_args=["AppGlobalNumByteSlice"],
            args=[app],
        )

    @classmethod
    def localNumUint(cls, app: Expr) -> MaybeValue:
        """Get the number of uint64 values allowed in Local State for the application.

        Args:
            app: An index into Txn.applications that correspond to the application to check.
                Must evaluate to uint64.
        """
        require_type(app, TealType.uint64)
        return MaybeValue(
            Op.app_params_get,
            TealType.uint64,
            immediate_args=["AppLocalNumUint"],
            args=[app],
        )

    @classmethod
    def localNumByteSlice(cls, app: Expr) -> MaybeValue:
        """Get the number of byte array values allowed in Local State for the application.

        Args:
            app: An index into Txn.applications that correspond to the application to check.
                Must evaluate to uint64.
        """
        require_type(app, TealType.uint64)
        return MaybeValue(
            Op.app_params_get,
            TealType.uint64,
            immediate_args=["AppLocalNumByteSlice"],
            args=[app],
        )

    @classmethod
    def extraProgramPages(cls, app: Expr) -> MaybeValue:
        """Get the number of Extra Program Pages of code space for the application.

        Args:
            app: An index into Txn.applications that correspond to the application to check.
                Must evaluate to uint64.
        """
        require_type(app, TealType.uint64)
        return MaybeValue(
            Op.app_params_get,
            TealType.uint64,
            immediate_args=["AppExtraProgramPages"],
            args=[app],
        )

    @classmethod
    def creator(cls, app: Expr) -> MaybeValue:
        """Get the creator address for the application.

        Args:
            app: An index into Txn.applications that correspond to the application to check.
                Must evaluate to uint64.
        """
        require_type(app, TealType.uint64)
        return MaybeValue(
            Op.app_params_get, TealType.bytes, immediate_args=["AppCreator"], args=[app]
        )

    @classmethod
    def address(cls, app: Expr) -> MaybeValue:
        """Get the escrow address for the application.

        Args:
            app: An index into Txn.applications that correspond to the application to check.
                Must evaluate to uint64.
        """
        require_type(app, TealType.uint64)
        return MaybeValue(
            Op.app_params_get, TealType.bytes, immediate_args=["AppAddress"], args=[app]
        )


AppParam.__module__ = "pyteal"


class AppParamObject:
    """Represents information about an application's parameters"""

    def __init__(self, app: Expr) -> None:
        """Create a new AppParamObject for the given application.

        Args:
            app: An identifier for the app. It must be an index into Txn.ForeignApps that
                corresponds to the app to check, or since v4, an application ID that appears in
                Txn.ForeignApps or is the CurrentApplicationID. In either case, it must evaluate to
                uint64.
        """
        require_type(app, TealType.uint64)
        self._app: Final = app

    def approval_program(self) -> MaybeValue:
        """Get the bytecode of Approval Program for the application."""
        return AppParam.approvalProgram(self._app)

    def clear_state_program(self) -> MaybeValue:
        return AppParam.clearStateProgram(self._app)

    def global_num_uint(self) -> MaybeValue:
        """Get the number of uint64 values allowed in Global State for the application."""
        return AppParam.globalNumUint(self._app)

    def global_num_byte_slice(self) -> MaybeValue:
        """Get the number of byte array values allowed in Global State for the application."""
        return AppParam.globalNumByteSlice(self._app)

    def local_num_uint(self) -> MaybeValue:
        """Get the number of uint64 values allowed in Local State for the application."""
        return AppParam.localNumUint(self._app)

    def local_num_byte_slice(self) -> MaybeValue:
        """Get the number of byte array values allowed in Local State for the application."""
        return AppParam.localNumByteSlice(self._app)

    def extra_program_pages(self) -> MaybeValue:
        """Get the number of Extra Program Pages of code space for the application."""
        return AppParam.extraProgramPages(self._app)

    def creator_address(self) -> MaybeValue:
        """Get the creator address for the application."""
        return AppParam.creator(self._app)

    def address(self) -> MaybeValue:
        """Get the escrow address for the application."""
        return AppParam.address(self._app)


AppParamObject.__module__ = "pyteal"



import pytest

import pyteal as pt

options = pt.CompileOptions()


def test_scratchvar_type():
    myvar_default = pt.ScratchVar()
    assert myvar_default.storage_type() == pt.TealType.anytype
    assert myvar_default.store(pt.Bytes("value")).type_of() == pt.TealType.none
    assert myvar_default.load().type_of() == pt.TealType.anytype

    with pytest.raises(pt.TealTypeError):
        myvar_default.store(pt.Pop(pt.Int(1)))

    myvar_int = pt.ScratchVar(pt.TealType.uint64)
    assert myvar_int.storage_type() == pt.TealType.uint64
    assert myvar_int.store(pt.Int(1)).type_of() == pt.TealType.none
    assert myvar_int.load().type_of() == pt.TealType.uint64

    with pytest.raises(pt.TealTypeError):
        myvar_int.store(pt.Bytes("value"))

    with pytest.raises(pt.TealTypeError):
        myvar_int.store(pt.Pop(pt.Int(1)))

    myvar_bytes = pt.ScratchVar(pt.TealType.bytes)
    assert myvar_bytes.storage_type() == pt.TealType.bytes
    assert myvar_bytes.store(pt.Bytes("value")).type_of() == pt.TealType.none
    assert myvar_bytes.load().type_of() == pt.TealType.bytes

    with pytest.raises(pt.TealTypeError):
        myvar_bytes.store(pt.Int(0))

    with pytest.raises(pt.TealTypeError):
        myvar_bytes.store(pt.Pop(pt.Int(1)))


def test_scratchvar_store():
    myvar = pt.ScratchVar(pt.TealType.bytes)
    arg = pt.Bytes("value")
    expr = myvar.store(arg)

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(arg, pt.Op.byte, '"value"'),
            pt.TealOp(expr, pt.Op.store, myvar.slot),
        ]
    )

    actual, _ = expr.__teal__(options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected


def test_scratchvar_load():
    myvar = pt.ScratchVar()
    expr = myvar.load()

    expected = pt.TealSimpleBlock([pt.TealOp(expr, pt.Op.load, myvar.slot)])

    actual, _ = expr.__teal__(options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected


def test_scratchvar_index():
    myvar = pt.ScratchVar()
    expr = myvar.index()

    expected = pt.TealSimpleBlock([pt.TealOp(expr, pt.Op.int, myvar.slot)])

    actual, _ = expr.__teal__(options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected


def test_scratchvar_assign_store():
    slotId = 2
    myvar = pt.ScratchVar(pt.TealType.uint64, slotId)
    arg = pt.Int(10)
    expr = myvar.store(arg)

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(arg, pt.Op.int, 10),
            pt.TealOp(expr, pt.Op.store, myvar.slot),
        ]
    )

    actual, _ = expr.__teal__(options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected


def test_scratchvar_assign_load():
    slotId = 5
    myvar = pt.ScratchVar(slotId=slotId)
    expr = myvar.load()

    expected = pt.TealSimpleBlock([pt.TealOp(expr, pt.Op.load, myvar.slot)])

    actual, _ = expr.__teal__(options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected


def test_dynamic_scratchvar_type():
    myvar_default = pt.DynamicScratchVar()
    assert myvar_default.storage_type() == pt.TealType.anytype
    assert myvar_default.store(pt.Bytes("value")).type_of() == pt.TealType.none
    assert myvar_default.load().type_of() == pt.TealType.anytype

    with pytest.raises(pt.TealTypeError):
        myvar_default.store(pt.Pop(pt.Int(1)))

    myvar_int = pt.DynamicScratchVar(pt.TealType.uint64)
    assert myvar_int.storage_type() == pt.TealType.uint64
    assert myvar_int.store(pt.Int(1)).type_of() == pt.TealType.none
    assert myvar_int.load().type_of() == pt.TealType.uint64

    with pytest.raises(pt.TealTypeError):
        myvar_int.store(pt.Bytes("value"))

    with pytest.raises(pt.TealTypeError):
        myvar_int.store(pt.Pop(pt.Int(1)))

    myvar_bytes = pt.DynamicScratchVar(pt.TealType.bytes)
    assert myvar_bytes.storage_type() == pt.TealType.bytes
    assert myvar_bytes.store(pt.Bytes("value")).type_of() == pt.TealType.none
    assert myvar_bytes.load().type_of() == pt.TealType.bytes

    with pytest.raises(pt.TealTypeError):
        myvar_bytes.store(pt.Int(0))

    with pytest.raises(pt.TealTypeError):
        myvar_bytes.store(pt.Pop(pt.Int(1)))


def test_dynamic_scratchvar_load():
    myvar = pt.DynamicScratchVar()
    expr = myvar.load()

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(pt.ScratchLoad(myvar.slot), pt.Op.load, myvar.slot),
            pt.TealOp(expr, pt.Op.loads),
        ]
    )

    actual, _ = expr.__teal__(options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def test_dynamic_scratchvar_store():
    myvar = pt.DynamicScratchVar(pt.TealType.bytes)
    arg = pt.Bytes("value")
    expr = myvar.store(arg)

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(pt.ScratchLoad(myvar.slot), pt.Op.load, myvar.slot),
            pt.TealOp(arg, pt.Op.byte, '"value"'),
            pt.TealOp(expr, pt.Op.stores),
        ]
    )

    actual, _ = expr.__teal__(options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def test_dynamic_scratchvar_index():
    myvar = pt.DynamicScratchVar()
    expr = myvar.index()

    expected = pt.TealSimpleBlock([pt.TealOp(expr, pt.Op.load, myvar.slot)])

    actual, _ = expr.__teal__(options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected


def test_dynamic_scratchvar_cannot_set_index_to_another_dynamic():
    myvar = pt.DynamicScratchVar()
    myvar.load()

    regvar = pt.ScratchVar()
    myvar.set_index(regvar)

    dynvar = pt.DynamicScratchVar()

    with pytest.raises(pt.TealInputError):
        myvar.set_index(dynvar)



from typing import Union, Sequence, TypeVar, cast

from pyteal.errors import TealInputError
from pyteal.ast.expr import Expr
from pyteal.ast.seq import Seq
from pyteal.ast.int import Int
from pyteal.ast.substring import Suffix

from pyteal.ast.abi.type import ComputedValue, BaseType
from pyteal.ast.abi.uint import Uint16, Byte, ByteTypeSpec
from pyteal.ast.abi.array_base import ArrayTypeSpec, Array


T = TypeVar("T", bound=BaseType)


class DynamicArrayTypeSpec(ArrayTypeSpec[T]):
    def new_instance(self) -> "DynamicArray[T]":
        return DynamicArray(self)

    def annotation_type(self) -> type["DynamicArray[T]"]:
        return DynamicArray[self.value_type_spec().annotation_type()]  # type: ignore[misc]

    def is_length_dynamic(self) -> bool:
        return True

    def is_dynamic(self) -> bool:
        return True

    def byte_length_static(self) -> int:
        raise ValueError("Type is dynamic")

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, DynamicArrayTypeSpec)
            and self.value_type_spec() == other.value_type_spec()
        )

    def __str__(self) -> str:
        return f"{self.value_type_spec()}[]"


DynamicArrayTypeSpec.__module__ = "pyteal.abi"


class DynamicArray(Array[T]):
    """The class that represents ABI dynamic array type."""

    def __init__(self, array_type_spec: DynamicArrayTypeSpec[T]) -> None:
        super().__init__(array_type_spec)

    def type_spec(self) -> DynamicArrayTypeSpec[T]:
        return cast(DynamicArrayTypeSpec[T], super().type_spec())

    def set(
        self,
        values: Union[Sequence[T], "DynamicArray[T]", ComputedValue["DynamicArray[T]"]],
    ) -> Expr:
        """
        Set the elements of this DynamicArray to the input values.

        The behavior of this method depends on the input argument type:

            * :code:`Sequence[T]`: set the elements of this DynamicArray to those contained in this Python sequence (e.g. a list or tuple). A compiler error will occur if any element in the sequence does not match this DynamicArray's element type.
            * :code:`DynamicArray[T]`: copy the elements from another DynamicArray. The argument's element type must exactly match this DynamicArray's element type, otherwise an error will occur.
            * :code:`ComputedValue[DynamicArray[T]]`: copy the elements from a DynamicArray produced by a ComputedValue. The element type produced by the ComputedValue must exactly match this DynamicArray's element type, otherwise an error will occur.

        Args:
            values: The new elements this DynamicArray should have. This must follow the above constraints.

        Returns:
            An expression which stores the given value into this DynamicArray.
        """

        if isinstance(values, ComputedValue):
            return self._set_with_computed_type(values)
        elif isinstance(values, BaseType):
            if self.type_spec() != values.type_spec():
                raise TealInputError(
                    f"Cannot assign type {values.type_spec()} to {self.type_spec()}"
                )
            return self.stored_value.store(values.encode())
        return super().set(values)

    def length(self) -> Expr:
        """Get the element number of this ABI dynamic array.

        The array length (element number) is encoded in the first 2 bytes of the byte encoding.

        Returns:
            A PyTeal expression that represents the dynamic array length.
        """
        output = Uint16()
        return Seq(
            output.decode(self.encode()),
            output.get(),
        )


DynamicArray.__module__ = "pyteal.abi"


class DynamicBytesTypeSpec(DynamicArrayTypeSpec[Byte]):
    def __init__(self) -> None:
        super().__init__(ByteTypeSpec())

    def new_instance(self) -> "DynamicBytes":
        return DynamicBytes()

    def annotation_type(self) -> type["DynamicBytes"]:
        return DynamicBytes

    def __str__(self) -> str:
        return "byte[]"

    def __eq__(self, other: object) -> bool:
        return isinstance(other, DynamicBytesTypeSpec)


DynamicBytesTypeSpec.__module__ = "pyteal.abi"


class DynamicBytes(DynamicArray[Byte]):
    """The convenience class that represents ABI dynamic byte array."""

    def __init__(self) -> None:
        super().__init__(DynamicBytesTypeSpec())

    def type_spec(self) -> DynamicBytesTypeSpec:
        return DynamicBytesTypeSpec()

    def set(
        self,
        values: Union[
            bytes,
            bytearray,
            Expr,
            Sequence[Byte],
            DynamicArray[Byte],
            ComputedValue[DynamicArray[Byte]],
        ],
    ) -> Expr:
        """Set the elements of this DynamicBytes to the input values.

        The behavior of this method depends on the input argument type:

            * :code:`bytes`: set the value to the Python byte string.
            * :code:`bytearray`: set the value to the Python byte array.
            * :code:`Expr`: set the value to the result of a PyTeal expression, which must evaluate to a TealType.bytes.
            * :code:`Sequence[Byte]`: set the bytes of this String to those contained in this Python sequence (e.g. a list or tuple).
            * :code:`DynamicArray[Byte]`: copy the bytes from another DynamicArray. The argument's element type must exactly match Byte, otherwise an error will occur.
            * :code:`ComputedValue[DynamicArray[Byte]]`: copy the bytes from a DynamicArray produced by a ComputedValue. The argument's element type must exactly match Byte, otherwise an error will occur.

        Args:
            values: The new elements this DynamicBytes should have. This must follow the above constraints.

        Returns:
            An expression which stores the given value into this DynamicBytes.
        """
        # NOTE: the import here is to avoid importing in partial initialized module abi
        from pyteal.ast.abi.string import (
            _encoded_byte_string,
            _store_encoded_expr_byte_string_into_var,
        )

        match values:
            case bytes() | bytearray():
                return self.stored_value.store(_encoded_byte_string(values))
            case Expr():
                return _store_encoded_expr_byte_string_into_var(
                    values, self.stored_value
                )

        return super().set(values)

    def get(self) -> Expr:
        """Get the byte encoding of this DynamicBytes.

        Dropping the uint16 encoding prefix for dynamic array length.

        Returns:
            A Pyteal expression that loads byte encoding of this DynamicBytes, and drop the first uint16 DynamicArray length encoding.
        """
        return Suffix(self.stored_value.load(), Int(2))


DynamicBytes.__module__ = "pyteal.abi"



from pyteal.ast import *
from pyteal.ast import __all__ as ast_all
from pyteal.pragma import pragma
from pyteal.ir import *
from pyteal.ir import __all__ as ir_all
from pyteal.compiler import (
    MAX_TEAL_VERSION,
    MIN_TEAL_VERSION,
    DEFAULT_TEAL_VERSION,
    MAX_PROGRAM_VERSION,
    MIN_PROGRAM_VERSION,
    DEFAULT_PROGRAM_VERSION,
    CompileOptions,
    compileTeal,
    OptimizeOptions,
)
from pyteal.types import TealType
from pyteal.errors import (
    TealInternalError,
    TealTypeError,
    TealInputError,
    TealCompileError,
    TealPragmaError,
)
from pyteal.config import (
    MAX_GROUP_SIZE,
    NUM_SLOTS,
    RETURN_HASH_PREFIX,
    METHOD_ARG_NUM_CUTOFF,
)

# begin __all__
__all__ = (
    ast_all
    + ir_all
    + [
        "MAX_TEAL_VERSION",
        "MIN_TEAL_VERSION",
        "DEFAULT_TEAL_VERSION",
        "MAX_PROGRAM_VERSION",
        "MIN_PROGRAM_VERSION",
        "DEFAULT_PROGRAM_VERSION",
        "CompileOptions",
        "pragma",
        "compileTeal",
        "OptimizeOptions",
        "TealType",
        "TealInternalError",
        "TealTypeError",
        "TealInputError",
        "TealCompileError",
        "TealPragmaError",
        "MAX_GROUP_SIZE",
        "NUM_SLOTS",
        "RETURN_HASH_PREFIX",
        "METHOD_ARG_NUM_CUTOFF",
    ]
)
# end __all__



from enum import Enum
from typing import Tuple, TYPE_CHECKING

from pyteal.ast import Expr, MultiValue
from pyteal.errors import (
    TealTypeError,
    verifyFieldVersion,
    verifyProgramVersion,
    TealInputError,
)
from pyteal.ir import Op, TealBlock, TealOp
from pyteal.types import TealType, require_type

if TYPE_CHECKING:
    from pyteal.compiler import CompileOptions


EcdsaPubkey = [TealType.bytes, TealType.bytes]


class EcdsaCurve(Enum):
    """Enum representing an elliptic curve specification used in ECDSA."""

    Secp256k1 = (0, "Secp256k1", 5)
    Secp256r1 = (1, "Secp256r1", 7)

    def __init__(self, id: int, name: str, min_version: int) -> None:
        self.id = id
        self.arg_name = name
        self.min_version = min_version


class EcdsaVerifyExpr(Expr):
    """Implements basic ECDSA verification functionality

    This class shouldn't be directly used in a PyTeal program. The EcdsaVerify function
    should be used instead."""

    def __init__(
        self,
        curve: EcdsaCurve,
        data: Expr,
        sigA: Expr,
        sigB: Expr,
        pkX: Expr,
        pkY: Expr,
    ):
        super().__init__()
        require_type(data, TealType.bytes)
        require_type(sigA, TealType.bytes)
        require_type(sigB, TealType.bytes)
        require_type(pkX, TealType.bytes)
        require_type(pkY, TealType.bytes)

        self.op = Op.ecdsa_verify
        self.curve = curve
        self.args = [data, sigA, sigB, pkX, pkY]

    def __teal__(self, options: "CompileOptions"):
        verifyProgramVersion(
            self.op.min_version,
            options.version,
            "Program version too low to use op {}".format(self.op),
        )

        verifyFieldVersion(self.curve.arg_name, self.curve.min_version, options.version)

        return TealBlock.FromOp(
            options, TealOp(self, self.op, self.curve.arg_name), *self.args
        )

    def __str__(self):
        return "({} {} {} {} {} {} {})".format(self.op, self.curve, *self.args)

    def type_of(self):
        return TealType.uint64

    def has_return(self):
        return False


def EcdsaVerify(
    curve: EcdsaCurve,
    data: Expr,
    sigA: Expr,
    sigB: Expr,
    pubkey: Tuple[Expr, Expr] | MultiValue,
) -> Expr:
    """Verify an ECDSA signature.

    The public key argument can be represented as either a tuple of Byte expressions
    representing the (X,Y) point on the elliptic curve or a MultiValue expression returning
    two Byte expressions, ex. the value returned by EcdsaDecompress or EcdsaRecover. All byte
    arguments must be big endian encoded.

    Args:
        curve: Enum representing the ECDSA curve used for the signature and public key
        data: Hash value of the signed data. Must be 32 bytes long.
        sigA: First component of the signature. Must evaluate to bytes.
        sigB: Second component of the signature. Must evaluate to bytes.
        pubkey: Public key used to verify signature. Represented as either a tuple of expressions
            that must evaluate to bytes or as a MultiValue expression that returns two byte values.

    Returns:
        An expression evaluating to either 0 or 1 representing the success of verification
    """

    if not isinstance(curve, EcdsaCurve):
        raise TealTypeError(curve, EcdsaCurve)

    if isinstance(pubkey, MultiValue):
        if pubkey.types != EcdsaPubkey:
            raise TealTypeError(pubkey.types, EcdsaPubkey)

        return pubkey.outputReducer(
            lambda X, Y: EcdsaVerifyExpr(curve, data, sigA, sigB, X, Y)
        )

    return EcdsaVerifyExpr(curve, data, sigA, sigB, pubkey[0], pubkey[1])


def EcdsaDecompress(curve: EcdsaCurve, compressed_pk: Expr) -> MultiValue:
    """Decompress an ECDSA public key.
    Args:
        curve: Enum representing the ECDSA curve used for the public key
        compressed_pk: The compressed public key. Must be 33 bytes long and big endian encoded.
    Returns:
        A MultiValue expression representing the two components of the public key, big endian
        encoded.
    """

    if not isinstance(curve, EcdsaCurve):
        raise TealTypeError(curve, EcdsaCurve)

    require_type(compressed_pk, TealType.bytes)
    return MultiValue(
        Op.ecdsa_pk_decompress,
        EcdsaPubkey,
        immediate_args=[curve.arg_name],
        args=[compressed_pk],
        compile_check=lambda options: verifyFieldVersion(
            curve.arg_name, curve.min_version, options.version
        ),
    )


def EcdsaRecover(
    curve: EcdsaCurve, data: Expr, recovery_id: Expr, sigA: Expr, sigB: Expr
) -> MultiValue:
    """Recover an ECDSA public key from a signature.
    All byte arguments must be big endian encoded.
    Args:
        curve: Enum representing the ECDSA curve used for the public key
        data: Hash value of the signed data. Must be 32 bytes long.
        recovery_id: value used to extract public key from signature. Must evaluate to uint.
        sigA: First component of the signature. Must evaluate to bytes.
        sigB: Second component of the signature. Must evaluate to bytes.
    Returns:
        A MultiValue expression representing the two components of the public key, big endian
        encoded.
    """

    if not isinstance(curve, EcdsaCurve):
        raise TealTypeError(curve, EcdsaCurve)

    if curve != EcdsaCurve.Secp256k1:
        raise TealInputError("Recover only supports Secp256k1")

    require_type(data, TealType.bytes)
    require_type(recovery_id, TealType.uint64)
    require_type(sigA, TealType.bytes)
    require_type(sigB, TealType.bytes)
    return MultiValue(
        Op.ecdsa_pk_recover,
        EcdsaPubkey,
        immediate_args=[curve.arg_name],
        args=[data, recovery_id, sigA, sigB],
        compile_check=lambda options: verifyFieldVersion(
            curve.arg_name, curve.min_version, options.version
        ),
    )



import pyteal as pt


def test_EqualityContext():
    expr1 = pt.Int(1)
    expr2 = pt.Int(1)

    op1 = pt.TealOp(expr1, pt.Op.int, 1)
    op2 = pt.TealOp(expr2, pt.Op.int, 1)

    assert op1 == op1
    assert op2 == op2
    assert op1 != op2
    assert op2 != op1

    with pt.TealComponent.Context.ignoreExprEquality():
        assert op1 == op1
        assert op2 == op2
        assert op1 == op2
        assert op2 == op1



from typing import TYPE_CHECKING
from enum import Enum

from pyteal.types import TealType, require_type
from pyteal.errors import verifyFieldVersion, verifyProgramVersion
from pyteal.ir import TealOp, Op, TealBlock
from pyteal.ast.expr import Expr

if TYPE_CHECKING:
    from pyteal.compiler import CompileOptions


class JsonRefType(Enum):
    # fmt: off
    #              id |    name   |      type      | min version
    string = (0, "JSONString", TealType.bytes,  7)
    uint64 = (1, "JSONUint64", TealType.uint64, 7)
    object = (2, "JSONObject", TealType.bytes,  7)
    # fmt: on

    def __init__(self, id: int, name: str, type: TealType, min_version: int) -> None:
        self.id = id
        self.arg_name = name
        self.ret_type = type
        self.min_version = min_version

    def type_of(self) -> TealType:
        return self.ret_type


JsonRefType.__module__ = "pyteal"


class JsonRef(Expr):
    """An expression that accesses the value associated with a given key from a supported utf-8 encoded json object.

    The json object must satisfy a `particular specification <https://github.com/algorand/go-algorand/blob/master/data/transactions/logic/jsonspec.md>`_.
    """

    def __init__(self, type: JsonRefType, json_obj: Expr, key: Expr) -> None:
        super().__init__()

        self.type = type

        require_type(json_obj, TealType.bytes)
        self.json_obj = json_obj

        require_type(key, TealType.bytes)
        self.key = key

    def __teal__(self, options: "CompileOptions"):
        verifyProgramVersion(
            Op.json_ref.min_version,
            options.version,
            "Program version too low to use op json_ref",
        )

        verifyFieldVersion(self.type.arg_name, self.type.min_version, options.version)

        op = TealOp(self, Op.json_ref, self.type.arg_name)
        return TealBlock.FromOp(options, op, self.json_obj, self.key)

    def __str__(self):
        return "(JsonRef {} {} {})".format(self.type.arg_name, self.json_obj, self.key)

    def type_of(self):
        return self.type.type_of()

    def has_return(self):
        return False

    @classmethod
    def as_string(cls, json_obj: Expr, key: Expr) -> Expr:
        """Access the value of a given key as a string.

        Refer to the `JsonRef` class documentation for valid json specification.

        Args:
            json_obj: The utf-8 encoded json object.
            key: The key to access in the json object.
        """
        return cls(JsonRefType.string, json_obj, key)

    @classmethod
    def as_uint64(cls, json_obj: Expr, key: Expr) -> Expr:
        """Access the value of a given key as a uint64.

        Refer to the `JsonRef` class documentation for valid json specification.

        Args:
            json_obj: The utf-8 encoded json object.
            key: The key to access in the json object.
        """
        return cls(JsonRefType.uint64, json_obj, key)

    @classmethod
    def as_object(cls, json_obj: Expr, key: Expr) -> "JsonRef":
        """Access the value of a given key as a json object.

        Refer to the `JsonRef` class documentation for valid json specification.

        Args:
            json_obj: The utf-8 encoded json object.
            key: The key to access in the json object.
        """
        return cls(JsonRefType.object, json_obj, key)


JsonRef.__module__ = "pyteal"



from typing import Union, cast, TYPE_CHECKING

from pyteal.types import TealType, require_type
from pyteal.ir import TealOp, Op, TealBlock
from pyteal.errors import TealInputError, verifyFieldVersion, verifyProgramVersion
from pyteal.config import MAX_GROUP_SIZE
from pyteal.ast.expr import Expr
from pyteal.ast.txn import TxnField, TxnExpr, TxnaExpr, TxnObject

if TYPE_CHECKING:
    from pyteal.compiler import CompileOptions


def validate_txn_index_or_throw(txnIndex: Union[int, Expr]):
    if not isinstance(txnIndex, (int, Expr)):
        raise TealInputError(
            f"Invalid txnIndex type:  Expected int or Expr, but received {txnIndex}"
        )
    if isinstance(txnIndex, Expr):
        require_type(txnIndex, TealType.uint64)


class GtxnExpr(TxnExpr):
    """An expression that accesses a transaction field from a transaction in the current group."""

    def __init__(self, txnIndex: Union[int, Expr], field: TxnField) -> None:
        super().__init__(Op.gtxn, "Gtxn", field)
        validate_txn_index_or_throw(txnIndex)
        self.txnIndex = txnIndex

    def __str__(self):
        return "({} {} {})".format(self.name, self.txnIndex, self.field.arg_name)

    def __teal__(self, options: "CompileOptions"):
        verifyFieldVersion(self.field.arg_name, self.field.min_version, options.version)

        if type(self.txnIndex) is int:
            verifyProgramVersion(
                Op.gtxn.min_version,
                options.version,
                "Program version too low to use gtxn",
            )
            op = TealOp(self, Op.gtxn, self.txnIndex, self.field.arg_name)
            return TealBlock.FromOp(options, op)

        verifyProgramVersion(
            Op.gtxns.min_version,
            options.version,
            "Program version too low to index Gtxn with dynamic values",
        )

        op = TealOp(self, Op.gtxns, self.field.arg_name)
        return TealBlock.FromOp(options, op, cast(Expr, self.txnIndex))


GtxnExpr.__module__ = "pyteal"


class GtxnaExpr(TxnaExpr):
    """An expression that accesses a transaction array field from a transaction in the current group."""

    def __init__(
        self, txnIndex: Union[int, Expr], field: TxnField, index: Union[int, Expr]
    ) -> None:
        super().__init__(Op.gtxna, Op.gtxnas, "Gtxna", field, index)
        validate_txn_index_or_throw(txnIndex)
        self.txnIndex = txnIndex

    def __str__(self):
        return "({} {} {} {})".format(
            self.name, self.txnIndex, self.field.arg_name, self.index
        )

    def __teal__(self, options: "CompileOptions"):
        verifyFieldVersion(self.field.arg_name, self.field.min_version, options.version)

        if type(self.txnIndex) is int:
            if type(self.index) is int:
                opToUse = Op.gtxna
            else:
                opToUse = Op.gtxnas
        else:
            if type(self.index) is int:
                opToUse = Op.gtxnsa
            else:
                opToUse = Op.gtxnsas

        verifyProgramVersion(
            opToUse.min_version,
            options.version,
            "Program version too low to use op {}".format(opToUse),
        )

        if type(self.txnIndex) is int:
            if type(self.index) is int:
                op = TealOp(
                    self, opToUse, self.txnIndex, self.field.arg_name, self.index
                )
                return TealBlock.FromOp(options, op)
            op = TealOp(self, opToUse, self.txnIndex, self.field.arg_name)
            return TealBlock.FromOp(options, op, cast(Expr, self.index))

        if type(self.index) is int:
            op = TealOp(self, opToUse, self.field.arg_name, self.index)
            return TealBlock.FromOp(options, op, cast(Expr, self.txnIndex))

        op = TealOp(self, opToUse, self.field.arg_name)
        return TealBlock.FromOp(
            options, op, cast(Expr, self.txnIndex), cast(Expr, self.index)
        )


GtxnaExpr.__module__ = "pyteal"


class TxnGroup:
    """Represents a group of transactions."""

    def __getitem__(self, txnIndex: Union[int, Expr]) -> TxnObject:
        if type(txnIndex) is int:
            if txnIndex < 0 or txnIndex >= MAX_GROUP_SIZE:
                raise TealInputError(
                    "Invalid Gtxn index {}, shoud be in [0, {})".format(
                        txnIndex, MAX_GROUP_SIZE
                    )
                )
        else:
            require_type(cast(Expr, txnIndex), TealType.uint64)
        return TxnObject(
            lambda field: GtxnExpr(txnIndex, field),
            lambda field, index: GtxnaExpr(txnIndex, field, index),
        )


TxnGroup.__module__ = "pyteal"

Gtxn: TxnGroup = TxnGroup()

Gtxn.__module__ = "pyteal"



from algosdk.atomic_transaction_composer import ABI_RETURN_HASH


# Maximum size of an atomic transaction group.
MAX_GROUP_SIZE = 16

# Number of scratch space slots available.
NUM_SLOTS = 256

# Method return selector in base16
RETURN_HASH_PREFIX = ABI_RETURN_HASH

# Method argument number limit
METHOD_ARG_NUM_CUTOFF = 15



import pytest
import pkg_resources


@pytest.fixture
def mock_version(version: str, monkeypatch: pytest.MonkeyPatch):
    def mocked_require(name: str):
        if (
            name == "pyteal"
            and version is not None  # don't mock if no version is specified
        ):
            return [
                pkg_resources.Distribution(
                    version=version,
                )
            ]
        else:
            return pkg_resources.require(name)[0]

    monkeypatch.setattr(pkg_resources, "require", mocked_require)



# WARNING: this logic sig is for demo purposes only

from pyteal import (
    And,
    Arg,
    Btoi,
    Bytes,
    Expr,
    Global,
    If,
    Int,
    Pop,
    ScratchVar,
    Seq,
    Subroutine,
    TealType,
    Txn,
    TxnType,
)

ONE_ALGO = Int(1_000_000)


@Subroutine(TealType.uint64)
def root_closeness(A, B, C, X):
    left = ScratchVar(TealType.uint64)
    right = ScratchVar(TealType.uint64)
    return Seq(
        left.store(A * X * X + C),
        right.store(B * X),
        If(left.load() < right.load())
        .Then(right.load() - left.load())
        .Else(left.load() - right.load()),
    )


@Subroutine(TealType.uint64)
def calculate_prize(closeness):
    return (
        If(closeness + Int(1) < Int(20))
        .Then(ONE_ALGO * (Int(10) - (closeness + Int(1)) / Int(2)))
        .Else(Int(0))
    )


def logicsig(a: int, p: int, q: int) -> Expr:
    """
    Choices
    * (a, p, q) = (1, 5, 7)
    * compiling on program version 5 and
    * with assembleConstants = True
    results in Logic-Sig Contract Account Address:
    WO3TQD3WBSDKB6WEHUMSEBFH53GZVVXYGPWYDWKUZCKEXTVCDNDHJGG6II
    """
    assert all(
        isinstance(x, int) and p < q and a > 0 and x >= 0 for x in (a, p, q)
    ), f"require non-negative ints a, p, q with p < q but got {a, p, q}"

    b, c = a * (p + q), a * p * q
    msg = Bytes(f"Can you factor {a} * x^2 - {b} * x + {c} ?")

    A, B, C = Int(a), Int(b), Int(c)
    X1 = Btoi(Arg(0))
    X2 = Btoi(Arg(1))
    C1 = ScratchVar(TealType.uint64)
    C2 = ScratchVar(TealType.uint64)
    SUM = ScratchVar(TealType.uint64)
    PRIZE = ScratchVar(TealType.uint64)
    return Seq(
        Pop(msg),
        C1.store(root_closeness(A, B, C, X1)),
        C2.store(root_closeness(A, B, C, X2)),
        SUM.store(C1.load() + C2.load()),
        PRIZE.store(calculate_prize(SUM.load())),
        And(
            Txn.type_enum() == TxnType.Payment,
            Txn.close_remainder_to() == Global.zero_address(),
            X1 != X2,
            PRIZE.load(),
            Txn.amount() == PRIZE.load(),
        ),
    )


def create(a, b, c):
    return logicsig(*map(lambda x: int(x), (a, b, c)))



from typing import TYPE_CHECKING, Tuple

from pyteal.errors import TealInputError
from pyteal.types import TealType
from pyteal.ir import TealBlock, TealSimpleBlock, TealOp, Op
from pyteal.ast.expr import Expr
from pyteal.ast.seq import Seq

if TYPE_CHECKING:
    from pyteal.compiler import CompileOptions


class CommentExpr(Expr):
    """Represents a single line comment in TEAL source code.

    This class is intentionally hidden because it's too basic to directly expose. Anything exposed
    to users should be able to handle multi-line comments by breaking them apart and using this
    class.
    """

    def __init__(self, single_line_comment: str) -> None:
        super().__init__()
        if "\n" in single_line_comment or "\r" in single_line_comment:
            raise TealInputError(
                "Newlines should not be present in the CommentExpr constructor"
            )
        self.comment = single_line_comment

    def __teal__(self, options: "CompileOptions") -> Tuple[TealBlock, TealSimpleBlock]:
        op = TealOp(self, Op.comment, self.comment)
        return TealBlock.FromOp(options, op)

    def __str__(self):
        return f'(Comment "{self.comment}")'

    def type_of(self):
        return TealType.none

    def has_return(self):
        return False


CommentExpr.__module__ = "pyteal"


def Comment(comment: str, expr: Expr = None) -> Expr:
    """Wrap an existing expression with a comment.

    This comment will be present in the compiled TEAL source immediately before the first op of the
    expression.

    Note that when TEAL source is assembled into bytes, all comments are omitted.

    Args:
        comment: The comment that will be associated with the expression.
        expr: The expression to be commented.

    Returns:
        A new expression which is functionally equivalent to the input expression, but which will
        compile with the given comment string.
    """
    lines = comment.splitlines()
    comment_lines: list[Expr] = [CommentExpr(line) for line in lines]
    if expr is not None:
        comment_lines.append(expr)
    return Seq(*comment_lines)



from typing import TYPE_CHECKING

from pyteal.types import TealType
from pyteal.ir import TealOp, Op, TealBlock
from pyteal.ast.expr import Expr

if TYPE_CHECKING:
    from pyteal.compiler import CompileOptions


class Err(Expr):
    """Expression that causes the program to immediately fail when executed."""

    def __teal__(self, options: "CompileOptions"):
        op = TealOp(self, Op.err)
        return TealBlock.FromOp(options, op)

    def __str__(self):
        return "(err)"

    def type_of(self):
        return TealType.none

    def has_return(self):
        return True


Err.__module__ = "pyteal"



import re
import pkg_resources
from typing import Any
from semantic_version import Version, NpmSpec

from pyteal.errors import TealPragmaError


def __convert_pep440_compiler_version(compiler_version: str):
    """Convert PEP 440 version identifiers to valid NPM versions.

    For example:
        "1.0.0" -> "1.0.0"
        "1.0.0a1" -> "1.0.0-a1"
        "<0.5.0+local || >=1.0.0a9.post1.dev2" -> "<0.5.0 || >=1.0.0-alpha9.1.2"
    """
    NUMBER = r"(?:x|X|\*|0|[1-9][0-9]*)"
    LOCAL = r"[a-zA-Z0-9.]*"
    TRIM_PREFIX_RE = re.compile(
        r"""
            (?:v)?                                          # Strip optional initial v
            (?P<op><|<=|>=|>|=|\^|~|)                       # Operator, can be empty
            (?P<major>{nb})(?:\.(?P<minor>{nb})(?:\.(?P<patch>{nb}))?)?
            (?:(?P<prerel_type>a|b|rc)(?P<prerel>{nb}))?    # Optional pre-release
            (?:\.post(?P<postrel>{nb}))?                    # Optional post-release
            (?:\.dev(?P<dev>{nb}))?                         # Optional dev release
            (?:\+(?P<local>{lcl}))?                         # Optional local version
        """.format(
            nb=NUMBER,
            lcl=LOCAL,
        ),
        re.VERBOSE,
    )

    def match_replacer(match: re.Match):
        (
            op,
            major,
            minor,
            patch,
            prerel_type,
            prerel,
            postrel,
            dev,
            local,
        ) = match.groups()

        # Base version (major/minor/patch)
        base_version = "{}.{}.{}".format(major or "0", minor or "0", patch or "0")

        # Combine prerel, postrel, and dev
        combined_additions = []
        short_prerel_type_to_long = {
            "a": "alpha",
            "b": "beta",
            "rc": "rc",
        }
        if prerel_type is not None:
            combined_additions.append(short_prerel_type_to_long[prerel_type] + prerel)
        if len(combined_additions) > 0 or postrel is not None or dev is not None:
            combined_additions.append(postrel or "0")
        if len(combined_additions) > 0 or dev is not None:
            combined_additions.append(dev or "0")
        combined_additions_str = ".".join(combined_additions)

        # Build full_version
        full_version = base_version
        if len(combined_additions) > 0:
            full_version += "-" + combined_additions_str
        if local is not None:
            full_version += "+" + local.lower()

        if op is not None:
            return op + full_version
        return full_version

    return re.sub(TRIM_PREFIX_RE, match_replacer, compiler_version)


def is_valid_compiler_version(compiler_version: str):
    """Check if the compiler version is valid.

    Args:
        compiler_version: The compiler version to check.

    Returns:
        True if the compiler version is a valid NPM specification range
        using either the PEP 440 or semantic version format, otherwise False.
    """
    try:
        pep440_converted = __convert_pep440_compiler_version(compiler_version)
        NpmSpec(pep440_converted)
        return True
    except ValueError:
        return False


def pragma(
    *,
    compiler_version: str,
    **kwargs: Any,
) -> None:
    """
    Specify pragmas for the compiler.

    Args:
        compiler_version: Acceptable versions of the compiler. Will fail if the current PyTeal version
            is not contained in the range. Follows the npm `semver range scheme <https://github.com/npm/node-semver#ranges>`_
            for specifying compatible versions.

    For example:

        .. code-block:: python

            # this will immediately fail if the current PyTeal version does not satisfy the
            # version constraint
            pragma(compiler_version="^0.14.0")
    """
    pkg_version = pkg_resources.require("pyteal")[0].version
    pyteal_version = Version(__convert_pep440_compiler_version(pkg_version))
    if pyteal_version not in NpmSpec(
        __convert_pep440_compiler_version(compiler_version)
    ):
        raise TealPragmaError(
            "PyTeal version {} is not compatible with compiler version {}".format(
                pkg_version, compiler_version
            )
        )



from pathlib import Path

from pyteal.compiler import compileTeal
from pyteal.ir import Mode

PATH = Path.cwd() / "tests" / "unit"
FIXTURES = PATH / "teal"
GENERATED = PATH / "generated"


def compile_and_save(approval, version, test_name):
    compiled = compileTeal(approval(), mode=Mode.Application, version=version)
    name = approval.__name__
    tealdir = GENERATED / test_name
    tealdir.mkdir(parents=True, exist_ok=True)
    with open(tealdir / (name + ".teal"), "w") as f:
        f.write(compiled)
    print(
        f"""Successfuly tested approval program <<{name}>> having 
compiled it into {len(compiled)} characters. See the results in:
{tealdir}
"""
    )
    return tealdir, name, compiled


def mismatch_ligature(expected, actual):
    la, le = len(actual), len(expected)
    mm_idx = -1
    for i in range(min(la, le)):
        if expected[i] != actual[i]:
            mm_idx = i
            break
    if mm_idx < 0:
        return ""
    return " " * (mm_idx) + "X" + "-" * (max(la, le) - mm_idx - 1)


def assert_teal_as_expected(path2actual, path2expected):
    with open(path2actual, "r") as fa, open(path2expected, "r") as fe:
        alines = fa.read().split("\n")
        elines = fe.read().split("\n")

        assert len(elines) == len(
            alines
        ), f"""EXPECTED {len(elines)} lines for {path2expected} but ACTUALLY got {len(alines)} lines in {path2actual}"""

        for i, actual in enumerate(alines):
            expected = elines[i]
            assert expected.startswith(
                actual
            ), f"""ACTUAL line in {path2actual}
LINE{i+1}:
{actual}
{mismatch_ligature(expected, actual)}
DOES NOT prefix the EXPECTED (which should have been actual + some commentary) in {path2expected}:
LINE{i+1}:
{expected}  
{mismatch_ligature(expected, actual)}
"""


def assert_new_v_old(approve_func, version, test_name):
    tealdir, name, compiled = compile_and_save(approve_func, version, test_name)

    print(
        f"""Compilation resulted in TEAL program of length {len(compiled)}. 
To view output SEE <{name}.teal> in ({tealdir})
--------------"""
    )

    path2actual = tealdir / (name + ".teal")
    path2expected = FIXTURES / test_name / (name + ".teal")
    assert_teal_as_expected(path2actual, path2expected)
